from rest_framework import status

from django.core import mail

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_email_sent
from ..permissions import (
    permission_mailing_profile_use,
    permission_send_document_file_attachment,
    permission_send_document_file_link, permission_send_document_link,
    permission_send_document_version_attachment,
    permission_send_document_version_link
)

from .literals import (
    TEST_EMAIL_ADDRESS, TEST_EMAIL_FROM_ADDRESS,
    TEST_MAILING_OBJECT_CONTENT, TEST_MAILING_OBJECT_MIME_TYPE
)
from .mixins import MailObjectSendAPIViewTestMixin


class MailDocumentFileSendAPIViewTestCase(
    DocumentTestMixin, MailObjectSendAPIViewTestMixin, BaseAPITestCase
):
    _test_mailing_profile_auto_create = True

    def setUp(self):
        super().setUp()

        self._test_object = self._test_document_file

        self._inject_test_object_content_type()

    def test_document_file_send_attachment_api_view_no_permission(self):
        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_attachment_api_view_with_mailing_profile_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_attachment_api_view_with_test_object_access(self):
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_attachment
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_attachment_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_attachment
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            len(
                mail.outbox[0].attachments
            ), 1
        )
        attachment = mail.outbox[0].attachments[0]
        self.assertEqual(
            attachment[0], str(self._test_object)
        )
        self.assertEqual(
            attachment[2], self._test_object.mimetype
        )

        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, [TEST_EMAIL_ADDRESS]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_file_send_attachment_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_attachment
        )

        self._test_object.document.delete()

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_api_view_no_permission(self):
        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_api_view_with_mailing_profile_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_api_view_with_test_object_access(self):
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_link
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_file_send_link_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_link
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            len(
                mail.outbox[0].attachments
            ), 0
        )
        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, [TEST_EMAIL_ADDRESS]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_file_send_link_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_link
        )

        self._test_object.document.delete()

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class MailDocumentVersionSendAPIViewTestCase(
    DocumentTestMixin, MailObjectSendAPIViewTestMixin, BaseAPITestCase
):
    _test_mailing_profile_auto_create = True

    def setUp(self):
        super().setUp()

        self._test_object = self._test_document_version

        self._inject_test_object_content_type()

    def test_document_version_send_attachment_api_view_no_permission(self):
        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_send_attachment_api_view_with_mailing_profile_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_send_attachment_api_view_with_test_object_access(self):
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_version_attachment
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_send_attachment_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_version_attachment
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            len(
                mail.outbox[0].attachments
            ), 1
        )
        attachment = mail.outbox[0].attachments[0]
        self.assertEqual(
            attachment[0], str(self._test_object)
        )
        self.assertEqual(
            attachment[2], 'application/pdf'
        )

        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, [TEST_EMAIL_ADDRESS]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_version_send_attachment_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_version_attachment
        )

        self._test_object.document.delete()

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_send_link_api_view_no_permission(self):
        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_send_link_api_view_with_mailing_profile_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_send_link_api_view_with_test_object_access(self):
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_version_link
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_version_send_link_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_version_link
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            len(
                mail.outbox[0].attachments
            ), 0
        )
        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, [TEST_EMAIL_ADDRESS]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_version_send_link_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_version_link
        )

        self._test_object.document.delete()

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class MailDocumentSendAPIViewTestCase(
    DocumentTestMixin, MailObjectSendAPIViewTestMixin, BaseAPITestCase
):
    _test_mailing_profile_auto_create = True

    def setUp(self):
        super().setUp()

        self._test_object = self._test_document

        self._inject_test_object_content_type()

    def test_document_send_attachment_api_view_no_permission(self):
        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_send_attachment_api_view_with_mailing_profile_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_send_link_api_view_no_permission(self):
        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_send_link_api_view_with_mailing_profile_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_send_link_api_view_with_test_object_access(self):
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_link
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_send_link_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_link
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            len(
                mail.outbox[0].attachments
            ), 0
        )
        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, [TEST_EMAIL_ADDRESS]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_trashed_document_send_link_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_link
        )

        self._test_object.delete()

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class MailTestObjectAPIViewTestCase(
    MailObjectSendAPIViewTestMixin, BaseAPITestCase
):
    _test_mailing_profile_auto_create = True
    _test_model_mailing_permission_attachment = permission_send_document_file_attachment
    _test_model_mailing_permission_link = permission_send_document_file_link
    auto_create_test_object = True

    def test_object_send_attachment_api_view_no_permission(self):
        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_object_send_attachment_api_view_with_mailing_profile_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_object_send_attachment_api_view_with_test_object_access(self):
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_attachment
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_object_send_attachment_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_attachment
        )

        self._clear_events()

        response = self._request_object_mail_attachment_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            len(
                mail.outbox[0].attachments
            ), 1
        )
        attachment = mail.outbox[0].attachments[0]
        self.assertEqual(
            attachment[0], str(self._test_object)
        )
        self.assertEqual(
            attachment[1], TEST_MAILING_OBJECT_CONTENT
        )
        self.assertEqual(
            attachment[2], TEST_MAILING_OBJECT_MIME_TYPE
        )

        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, [TEST_EMAIL_ADDRESS]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_email_sent.id)

    def test_object_send_link_api_view_no_permission(self):
        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_object_send_link_api_view_with_mailing_profile_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_object_send_link_api_view_with_test_object_access(self):
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_link
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            len(mail.outbox), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_object_send_link_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_mailing_profile,
            permission=permission_mailing_profile_use
        )
        self.grant_access(
            obj=self._test_object,
            permission=permission_send_document_file_link
        )

        self._clear_events()

        response = self._request_object_mail_link_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            len(mail.outbox), 1
        )
        self.assertEqual(
            len(
                mail.outbox[0].attachments
            ), 0
        )
        self.assertEqual(
            mail.outbox[0].from_email, TEST_EMAIL_FROM_ADDRESS
        )
        self.assertEqual(
            mail.outbox[0].to, [TEST_EMAIL_ADDRESS]
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_object)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_mailing_profile)
        self.assertEqual(events[0].verb, event_email_sent.id)



import logging

from django.apps import apps
from django.db.utils import OperationalError, ProgrammingError
from django.utils.functional import classproperty

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.utils import get_class_full_name

from .exceptions import FileMetadataError

logger = logging.getLogger(name=__name__)


class FileMetadataDriverCollection:
    _driver_enabled_list = []
    _driver_to_mime_type_dict = {}
    _mime_type_to_driver_dict = {}

    @classmethod
    def do_driver_disable(cls, driver):
        if driver in cls._driver_to_mime_type_dict:
            cls._driver_enabled_list.remove(driver)

    @classmethod
    def do_driver_disable_all(cls):
        cls._driver_enabled_list = []

    @classmethod
    def do_driver_enable(cls, driver):
        if driver in cls._driver_to_mime_type_dict:
            if driver not in cls._driver_enabled_list:
                cls._driver_enabled_list.append(driver)

    @classmethod
    def do_driver_register(cls, klass):
        if klass in cls._driver_to_mime_type_dict:
            raise FileMetadataError(
                'Driver "{}" is already registered.'.format(klass)
            )

        cls._driver_to_mime_type_dict[klass] = klass.mime_type_list

        for mime_type in klass.mime_type_list:
            cls._mime_type_to_driver_dict.setdefault(
                mime_type, []
            ).append(klass)

        klass.dotted_path = get_class_full_name(klass=klass)

        cls.do_driver_enable(driver=klass)

    @classmethod
    def get_all(cls, sorted=False):
        result = set()
        for mime_type, drivers in cls._mime_type_to_driver_dict.items():
            result.update(
                list(drivers)
            )

        result = list(result)

        if sorted:
            result.sort(key=lambda driver: driver.label)

        return result

    @classmethod
    def get_driver_for_mime_type(cls, mime_type):
        driver_class_list = cls._mime_type_to_driver_dict.get(
            mime_type, ()
        )
        # Add wildcard drivers, drivers meant to be executed for all MIME
        # types.
        driver_class_list += tuple(
            cls._mime_type_to_driver_dict.get(
                '*', ()
            )
        )

        result = [
            driver_class for driver_class in driver_class_list if driver_class in cls._driver_enabled_list
        ]

        return result


class FileMetadataDriverMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )

        if not new_class.__module__ == __name__:
            FileMetadataDriverCollection.do_driver_register(klass=new_class)

        return new_class


class FileMetadataDriver(
    AppsModuleLoaderMixin, metaclass=FileMetadataDriverMetaclass
):
    _loader_module_name = 'drivers'
    description = ''
    label = None
    internal_name = None
    mime_type_list = ()

    @classproperty
    def collection(cls):
        if cls != FileMetadataDriver:
            raise AttributeError(
                'This method is only available to the parent class.'
            )
        return FileMetadataDriverCollection

    @classmethod
    def do_model_instance_populate(cls):
        StoredDriver = apps.get_model(
            app_label='file_metadata', model_name='StoredDriver'
        )
        model_instance, created = StoredDriver.objects.get_or_create(
            driver_path=cls.dotted_path, defaults={
                'internal_name': cls.internal_name
            }
        )
        cls.model_instance = model_instance

    @classmethod
    def get_mime_type_list_display(cls):
        return ', '.join(cls.mime_type_list)

    @classmethod
    def post_load_modules(cls):
        StoredDriver = apps.get_model(
            app_label='file_metadata', model_name='StoredDriver'
        )

        try:
            """
            Check is the table is ready.
            If not, this will log an error similar to this:
            2023-12-12 09:12:46.985 UTC [79] ERROR:  relation "file_metadata_storeddriver" does not exist at character 22
            2023-12-12 09:12:46.985 UTC [79] STATEMENT:  SELECT 1 AS "a" FROM "file_metadata_storeddriver" LIMIT 1
            This error is expected and should be ignored.
            """
            StoredDriver.objects.exists()
        except (OperationalError, ProgrammingError):
            """
            This error is expected when trying to initialize the
            stored permissions during the initial creation of the
            database. Can be safely ignored under that situation.
            """
        else:
            for driver in cls.collection.get_all():
                driver.do_model_instance_populate()

    def __init__(self, auto_initialize=True, **kwargs):
        self.auto_initialize = auto_initialize

    def initialize(self):
        """
        Driver specific initialization code.
        """

    def process(self, document_file):
        logger.info(
            'Starting processing document file: %s', document_file
        )

        self.model_instance.driver_entries.filter(
            document_file=document_file
        ).delete()

        document_file_driver_entry = self.model_instance.driver_entries.create(
            document_file=document_file
        )

        results = self._process(document_file=document_file) or {}

        for key, value in results.items():
            document_file_driver_entry.entries.create(
                key=key, value=value
            )

    def _process(self, document_file):
        raise NotImplementedError(
            'Your %s class has not defined the required '
            '_process() method.' % self.__class__.__name__
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.common.serializers import ContentTypeSerializer
from mayan.apps.permissions.classes import Permission
from mayan.apps.permissions.serializers import RoleSerializer
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import (
    FilteredPrimaryKeyRelatedField, MultiKwargHyperlinkedIdentityField
)

from .models import AccessControlList


class ACLSerializer(serializers.ModelSerializer):
    content_type = ContentTypeSerializer(
        label=_(message='Content type'), read_only=True
    )
    permissions_add_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Permissions add URL'), view_kwargs=(
            {
                'lookup_field': 'content_type.app_label',
                'lookup_url_kwarg': 'app_label'
            },
            {
                'lookup_field': 'content_type.model',
                'lookup_url_kwarg': 'model_name'
            },
            {
                'lookup_field': 'object_id',
                'lookup_url_kwarg': 'object_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'acl_id'
            }
        ), view_name='rest_api:accesscontrollist-permission-add'
    )
    permissions_remove_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Permissions remove URL'), view_kwargs=(
            {
                'lookup_field': 'content_type.app_label',
                'lookup_url_kwarg': 'app_label'
            },
            {
                'lookup_field': 'content_type.model',
                'lookup_url_kwarg': 'model_name'
            },
            {
                'lookup_field': 'object_id',
                'lookup_url_kwarg': 'object_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'acl_id'
            }
        ), view_name='rest_api:accesscontrollist-permission-remove'
    )
    permissions_url = MultiKwargHyperlinkedIdentityField(
        label=_(message='Permissions URL'), view_kwargs=(
            {
                'lookup_field': 'content_type.app_label',
                'lookup_url_kwarg': 'app_label'
            },
            {
                'lookup_field': 'content_type.model',
                'lookup_url_kwarg': 'model_name'
            },
            {
                'lookup_field': 'object_id',
                'lookup_url_kwarg': 'object_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'acl_id'
            }
        ), view_name='rest_api:accesscontrollist-permission-list'
    )
    role = RoleSerializer(
        label=_(message='Role'), read_only=True
    )
    role_id = serializers.IntegerField(
        label=_(message='Role ID'), write_only=True
    )
    url = MultiKwargHyperlinkedIdentityField(
        label=_(message='URL'), view_kwargs=(
            {
                'lookup_field': 'content_type.app_label',
                'lookup_url_kwarg': 'app_label'
            },
            {
                'lookup_field': 'content_type.model',
                'lookup_url_kwarg': 'model_name'
            },
            {
                'lookup_field': 'object_id',
                'lookup_url_kwarg': 'object_id'
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'acl_id'
            }
        ), view_name='rest_api:accesscontrollist-detail'
    )

    class Meta:
        create_only_fields = ('role_id',)
        fields = (
            'content_type', 'id', 'object_id', 'permissions_add_url',
            'permissions_remove_url', 'permissions_url', 'role', 'role_id',
            'url'
        )
        model = AccessControlList
        read_only_fields = ('content_type', 'object_id')


class ACLPermissionAddSerializer(serializers.Serializer):
    permission = FilteredPrimaryKeyRelatedField(
        help_text=_(message='Primary key of the permission to add to the ACL.'),
        label=_(message='Permission ID'), source_queryset=Permission.all()
    )


class ACLPermissionRemoveSerializer(serializers.Serializer):
    permission = FilteredPrimaryKeyRelatedField(
        help_text=_(message='Primary key of the permission to remove from the ACL.'),
        label=_(message='Permission ID'), source_queryset=Permission.all()
    )



import logging

from PIL import Image, ImageColor, ImageDraw

from django import forms
from django.apps import apps
from django.utils.encoding import force_bytes
from django.utils.translation import gettext_lazy as _

from mayan.apps.views.forms import Form
from mayan.apps.views.widgets import ColorWidget

logger = logging.getLogger(name=__name__)


class ImagePasteTransformationMixin:
    class Form(Form):
        rotation = forms.IntegerField(
            help_text=_(
                'Number of degrees to rotate the image counter clockwise '
                'around its center.'
            ), label=_(message='Rotation'), required=False
        )
        transparency = forms.FloatField(
            help_text=_(message='Opacity level of the image in percent'),
            label=_(message='Transparency'), required=False
        )
        zoom = forms.FloatField(
            help_text=_(message='Zoom level in percent.'),
            label=_(message='Zoom'), required=False
        )

    @classmethod
    def get_arguments(cls):
        arguments = super().get_arguments() + (
            'rotation', 'transparency', 'zoom'
        )
        return arguments

    def _update_hash(self):
        result = super()._update_hash()
        instance = self.get_model_instance()
        # Add the image hash to the transformation hash. Ensures
        # that the content object image is updated if the image is
        # updated even if the transformation itself is not updated.
        result.update(
            force_bytes(
                s=instance.get_hash()
            )
        )

        return result

    def execute_on(self, *args, **kwargs):
        super().execute_on(*args, **kwargs)
        return self._execute_on(self, *args, **kwargs)

    def get_images(self):
        try:
            self.transparency = float(self.transparency or '100.0')
        except ValueError:
            self.transparency = 100

        if self.transparency < 0:
            self.transparency = 0
        elif self.transparency > 100:
            self.transparency = 100

        try:
            self.rotation = int(self.rotation or '0') % 360
        except ValueError:
            self.rotation = 0

        try:
            self.zoom = float(self.zoom or '100.0')
        except ValueError:
            self.zoom = 100.0

        instance = self.get_model_instance()

        instance_image = instance.get_image()

        if instance_image.mode != 'RGBA':
            instance_image.putalpha(alpha=255)

        instance_image = instance_image.rotate(
            angle=360 - self.rotation, resample=Image.Resampling.BICUBIC,
            expand=True
        )

        if self.zoom != 100.0:
            zoom_decimal_value = self.zoom / 100.0
            instance_image = instance_image.resize(
                size=(
                    int(instance_image.size[0] * zoom_decimal_value),
                    int(instance_image.size[1] * zoom_decimal_value)
                ), resample=Image.Resampling.LANCZOS
            )

        paste_mask = instance_image.getchannel(channel='A').point(
            lut=lambda pixel: int(pixel * self.transparency / 100)
        )

        return {
            'instance_image': instance_image, 'paste_mask': paste_mask
        }


class ImagePasteCoordinatesAbsoluteTransformationMixin(
    ImagePasteTransformationMixin
):
    arguments = ('left', 'top')
    label = _(message='Paste an image')
    name = 'paste_image'

    class Form(ImagePasteTransformationMixin.Form):
        left = forms.IntegerField(
            help_text=_(
                message='Horizontal position in pixels from the left.'
            ), label=_(message='Left'), required=False
        )
        top = forms.IntegerField(
            help_text=_(message='Vertical position in pixels from the top.'),
            label=_(message='Top'), required=False
        )

    def _execute_on(self, *args, **kwargs):
        try:
            self.left = int(self.left or '0')
        except ValueError:
            self.left = 0

        try:
            self.top = int(self.top or '0')
        except ValueError:
            self.top = 0

        if self.left < 0:
            self.left = 0

        if self.top < 0:
            self.top = 0

        images = self.get_images()

        self.image.paste(
            box=(self.left, self.top), im=images['instance_image'],
            mask=images['paste_mask']
        )

        return self.image


class ImagePasteCoordinatesPercentTransformationMixin(
    ImagePasteTransformationMixin
):
    arguments = ('left', 'top')
    label = _(message='Paste an image (percents coordinates)')
    name = 'paste_image_percent'

    class Form(ImagePasteTransformationMixin.Form):
        left = forms.FloatField(
            help_text=_(
                message='Horizontal position in percent from the left.'
            ), label=_(message='Left'), required=False
        )
        top = forms.FloatField(
            help_text=_(
                message='Vertical position in percent from the top.'
            ), label=_(message='Top'), required=False
        )

    def _execute_on(self, *args, **kwargs):
        try:
            self.left = float(self.left or '0')
        except ValueError:
            self.left = 0

        try:
            self.top = float(self.top or '0')
        except ValueError:
            self.top = 0

        if self.left < 0:
            self.left = 0

        if self.left > 100:
            self.left = 100

        if self.top < 0:
            self.top = 0

        if self.top > 100:
            self.top = 100

        images = self.get_images()

        base_width, base_height = self.image.size
        image_width, image_height = images['instance_image'].size

        self.left = int(
            self.left / 100.0 * (base_width - image_width)
        )
        self.top = int(
            self.top / 100.0 * (base_height - image_height)
        )

        self.image.paste(
            box=(self.left, self.top), im=images['instance_image'],
            mask=images['paste_mask']
        )

        return self.image


class ImageWatermarkPercentTransformationMixin(
    ImagePasteTransformationMixin
):
    arguments = (
        'left', 'top', 'right', 'bottom', 'horizontal_increment',
        'vertical_increment'
    )
    label = _(message='Paste an asset as watermark')
    name = 'paste_asset_watermark'

    class Form(ImagePasteTransformationMixin.Form):
        left = forms.IntegerField(
            help_text=_(
                'Horizontal start position in pixels from the left.'
            ), label=_(message='Left'), required=False
        )
        right = forms.IntegerField(
            help_text=_(
                message='Horizontal end position in pixels from the right.'
            ), label=_(message='Right'), required=False
        )
        top = forms.IntegerField(
            help_text=_(
                message='Vertical start position in pixels from the top.'
            ), label=_(message='Top'), required=False
        )
        bottom = forms.IntegerField(
            help_text=_(
                message='Vertical end position in pixels from the top.'
            ), label=_(message='Bottom'), required=False
        )
        horizontal_increment = forms.IntegerField(
            help_text=_(message='Horizontal position increments in pixels.'),
            label=_(message='Horizontal increment'), required=False
        )
        vertical_increment = forms.IntegerField(
            help_text=_(message='Vertical position increments in pixels.'),
            label=_(message='Vertical increment'), required=False
        )

    def _execute_on(self, *args, **kwargs):
        try:
            self.left = int(self.left or '0')
        except ValueError:
            self.left = 0

        try:
            self.top = int(self.top or '0')
        except ValueError:
            self.top = 0

        try:
            self.right = int(self.right or '0')
        except ValueError:
            self.right = 0

        try:
            self.bottom = int(self.bottom or '0')
        except ValueError:
            self.bottom = 0

        images = self.get_images()

        try:
            self.horizontal_increment = int(self.horizontal_increment or '0')
        except ValueError:
            self.horizontal_increment = 0

        try:
            self.vertical_increment = int(self.vertical_increment or '0')
        except ValueError:
            self.vertical_increment = 0

        if self.horizontal_increment == 0:
            self.horizontal_increment = images['paste_mask'].size[0]

        if self.vertical_increment == 0:
            self.vertical_increment = images['paste_mask'].size[1]

        for x in range(self.left, self.right or self.image.size[0], self.horizontal_increment):
            for y in range(self.top, self.bottom or self.image.size[1], self.vertical_increment):
                self.image.paste(
                    box=(x, y), im=images['instance_image'],
                    mask=images['paste_mask']
                )

        return self.image


class AssetTransformationMixin:
    @classmethod
    def get_form_class(cls):
        SuperForm = super().get_form_class()

        class FormWithSuperArguments(SuperForm):
            asset_name = forms.ChoiceField(
                help_text=_(message='Asset name'), label=_(message='Asset'),
                required=True
            )

            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

                def get_asset_choices():
                    queryset = apps.get_model(
                        app_label='converter', model_name='Asset'
                    ).objects.all()

                    for asset in queryset.all():
                        yield (asset.internal_name, asset)

                self.fields['asset_name'].choices = get_asset_choices()

        return FormWithSuperArguments

    @classmethod
    def get_arguments(cls):
        arguments = super().get_arguments() + (
            'asset_name',
        )
        return arguments

    def get_model_instance(self):
        asset_name = getattr(self, 'asset_name', None)

        Asset = apps.get_model(app_label='converter', model_name='Asset')

        try:
            asset = Asset.objects.get(internal_name=asset_name)
        except Asset.DoesNotExist:
            logger.error('Asset "%s" not found.', asset_name)
            raise
        else:
            return asset


class TransformationDrawRectangleMixin:
    class Form(Form):
        fillcolor = forms.CharField(
            help_text=_(message='Color used to fill the rectangle.'),
            label=_(message='Fill color'), required=False,
            widget=ColorWidget()
        )
        fill_transparency = forms.IntegerField(
            help_text=_(
                message='Opacity level of the fill color in percent'
            ), label=_(message='Fill transparency'), required=False
        )
        outlinecolor = forms.CharField(
            help_text=_(
                message='Color used for the outline of the rectangle.'
            ), label=_(message='Outline color'), required=False,
            widget=ColorWidget()
        )
        outlinewidth = forms.CharField(
            help_text=_(message='Width in pixels of the rectangle outline.'),
            label=_(message='Outline width'), required=False
        )

    def _execute_on(self, *args, **kwargs):
        fillcolor_value = getattr(self, 'fillcolor', None)
        if fillcolor_value:
            fill_color = ImageColor.getrgb(color=fillcolor_value)
        else:
            fill_color = (0, 0, 0)

        try:
            fill_transparency = int(
                getattr(self, 'fill_transparency', None) or '0'
            )
        except ValueError:
            fill_transparency = 100
        else:
            if fill_transparency < 0:
                fill_transparency = 0
            elif fill_transparency > 100:
                fill_transparency = 100

        # Convert transparency to opacity. Invert intensity logic, transpose
        # from percent to 8-bit value.
        opacity = int(
            (100 - fill_transparency) / 100 * 255
        )

        fill_color += (opacity,)

        outlinecolor_value = getattr(self, 'outlinecolor', None)
        if outlinecolor_value:
            outline_color = ImageColor.getrgb(color=outlinecolor_value)
        else:
            outline_color = None

        outlinewidth_value = getattr(self, 'outlinewidth', None)
        if outlinewidth_value:
            outline_width = int(outlinewidth_value)
        else:
            outline_width = 0

        draw = ImageDraw.Draw(im=self.image, mode='RGBA')
        draw.rectangle(
            fill=fill_color, outline=outline_color, width=outline_width,
            xy=(self.left, self.top, self.right, self.bottom)
        )

        return self.image



from django.utils.translation import gettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import menu_facet, menu_secondary, menu_tools
from mayan.apps.common.signals import (
    signal_post_initial_setup, signal_post_upgrade
)

from .handlers import (
    handler_search_backend_initialize, handler_search_backend_upgrade
)
from .links import (
    link_search, link_search_advanced, link_search_again,
    link_search_backend_reindex
)
from .search_backends import SearchBackend
from .search_models import SearchModel


class DynamicSearchApp(MayanAppConfig):
    app_namespace = 'search'
    app_url = 'search'
    has_rest_api = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.dynamic_search'
    verbose_name = _(message='Dynamic search')

    def ready(self):
        super().ready()

        SearchModel.load_modules()
        SearchBackend._enable()

        menu_facet.bind_links(
            links=(link_search, link_search_advanced),
            sources=(
                'search:search_simple', 'search:search_advanced',
                'search:search_results'
            )
        )
        menu_secondary.bind_links(
            links=(link_search_again,), sources=('search:search_results',)
        )
        menu_tools.bind_links(
            links=(link_search_backend_reindex,),
        )

        signal_post_initial_setup.connect(
            dispatch_uid='search_handler_search_backend_initialize',
            receiver=handler_search_backend_initialize
        )

        signal_post_upgrade.connect(
            dispatch_uid='search_handler_search_backend_upgrade',
            receiver=handler_search_backend_upgrade
        )



from mayan.apps.rest_api import generics

from .models import Key
from .permissions import (
    permission_key_delete, permission_key_upload, permission_key_view
)
from .serializers import KeySerializer


class APIKeyListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the keys.
    post: Upload a new key.
    """
    mayan_object_permission_map = {'GET': permission_key_view}
    mayan_view_permission_map = {'POST': permission_key_upload}
    serializer_class = KeySerializer
    source_queryset = Key.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APIKeyView(generics.RetrieveDestroyAPIView):
    """
    delete: Delete the selected key.
    get: Return the details of the selected key.
    """
    lookup_url_kwarg = 'key_id'
    mayan_object_permission_map = {
        'DELETE': permission_key_delete,
        'GET': permission_key_view
    }
    serializer_class = KeySerializer
    source_queryset = Key.objects.all()



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Mailing'), name='mailing'
)

# Mailing profile

permission_mailing_profile_create = namespace.add_permission(
    label=_(message='Create a mailing profile'),
    name='mailing_profile_create'
)
permission_mailing_profile_delete = namespace.add_permission(
    label=_(message='Delete a mailing profile'),
    name='mailing_profile_delete'
)
permission_mailing_profile_edit = namespace.add_permission(
    label=_(message='Edit a mailing profile'), name='mailing_profile_edit'
)
permission_mailing_profile_use = namespace.add_permission(
    label=_(message='Use a mailing profile'), name='mailing_profile_use'
)
permission_mailing_profile_view = namespace.add_permission(
    label=_(message='View a mailing profile'), name='mailing_profile_view'
)

# Document

permission_send_document_link = namespace.add_permission(
    label=_(message='Send document link via email'), name='mail_link'
)

# Document file

permission_send_document_file_attachment = namespace.add_permission(
    label=_(message='Send document file via email'),
    name='mail_document_file_attachment'
)
permission_send_document_file_link = namespace.add_permission(
    label=_(message='Send document file link via email'),
    name='mail_document_file_link'
)

# Document version

permission_send_document_version_attachment = namespace.add_permission(
    label=_(message='Send document version via email'),
    name='mail_document_version_attachment'
)
permission_send_document_version_link = namespace.add_permission(
    label=_(message='Send document version link via email'),
    name='mail_document_version_link'
)



from ..events import (
    event_document_trashed, event_trashed_document_deleted,
    event_trashed_document_restored
)
from ..models import Document, TrashedDocument

from .base import GenericDocumentTestCase


class TrashedDocumentModelTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_document_trashing(self):
        self._clear_events()

        # Trash the document.
        self._test_document.delete()
        self.assertEqual(
            TrashedDocument.objects.count(), 1
        )
        self.assertEqual(
            Document.valid.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_trashed.id)

    def test_trashed_document_restore(self):
        self._test_document.delete()

        self._clear_events()

        # Restore the document.
        test_trashed_document = TrashedDocument.objects.get(
            pk=self._test_document.pk
        )

        test_trashed_document.restore(user=self._test_case_user)

        self.assertEqual(
            TrashedDocument.objects.count(), 0
        )
        self.assertEqual(
            Document.valid.count(), 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_trashed_document_restored.id)

    def test_trashed_document_delete(self):
        self._test_document.delete()

        self._clear_events()

        # Delete the document.
        self._test_document.delete()
        self.assertEqual(
            TrashedDocument.objects.count(), 0
        )
        self.assertEqual(
            Document.valid.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_document_type)
        self.assertEqual(events[0].target, self._test_document_type)
        self.assertEqual(events[0].verb, event_trashed_document_deleted.id)


class TrashedDocumentPageModelTestCase(GenericDocumentTestCase):
    def test_trashed_document_page_count(self):
        page_count = self._test_document.version_active.pages.count()
        self._test_document.delete()
        test_trashed_document = TrashedDocument.objects.get(
            pk=self._test_document.pk
        )
        self.assertTrue(
            test_trashed_document.version_active.pages.count(), page_count
        )


class TrashedDocumentModelAPITestCase(GenericDocumentTestCase):
    def test_trashed_document_api_image_url(self):
        self._test_document.delete()
        test_trashed_document = TrashedDocument.objects.get(
            pk=self._test_document.pk
        )

        self._clear_events()

        self.assertTrue(
            test_trashed_document.get_api_image_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('acls', '0003_auto_20180402_0339')
    ]

    operations = [
        migrations.CreateModel(
            name='GlobalAccessControlListProxy',
            fields=[],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': []
            },
            bases=('acls.accesscontrollist',),
        ),
        migrations.AlterField(
            model_name='accesscontrollist',
            name='role',
            field=models.ForeignKey(
                help_text='Role to which the access is granted for the '
                'specified object.',
                on_delete=django.db.models.deletion.CASCADE,
                related_name='acls', to='permissions.Role',
                verbose_name='Role'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import icon_template_sandbox
from .permissions import permission_template_sandbox

link_document_template_sandbox = Link(
    args='resolved_object.pk', icon=icon_template_sandbox,
    permission=permission_template_sandbox, text=_(message='Sandbox'),
    view='templating:document_template_sandbox'
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Signature captures'),
    name='signature_captures'
)

event_signature_capture_created = namespace.add_event_type(
    label=_(message='Signature capture created'),
    name='signature_capture_created'
)
event_signature_capture_deleted = namespace.add_event_type(
    label=_(message='Signature capture deleted'),
    name='signature_capture_deleted'
)
event_signature_capture_edited = namespace.add_event_type(
    label=_(message='Signature capture edited'),
    name='signature_capture_edited'
)



import logging

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.template import RequestContext
from django.urls import reverse
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _

from mayan.apps.django_gpg.exceptions import NeedPassphrase, PassphraseError
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.views.generics import (
    ConfirmView, FormView, SingleObjectCreateView, SingleObjectDeleteView,
    SingleObjectDetailView, SingleObjectDownloadView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from .forms import (
    DocumentFileSignatureCreateForm, DocumentFileSignatureDetailForm
)
from .icons import (
    icon_document_file_all_signature_refresh,
    icon_document_file_all_signature_verify,
    icon_document_file_signature_detached_delete,
    icon_document_file_signature_detached_create,
    icon_document_file_signature_detail,
    icon_document_file_signature_detached_download,
    icon_document_file_signature_detached_upload,
    icon_document_file_signature_embedded_create,
    icon_document_file_signature_list
)
from .links import (
    link_document_file_signature_detached_create,
    link_document_file_signature_embedded_create,
    link_document_file_signature_detached_upload
)
from .models import DetachedSignature, EmbeddedSignature, SignatureBaseModel
from .permissions import (
    permission_document_file_sign_detached,
    permission_document_file_sign_embedded,
    permission_document_file_signature_delete,
    permission_document_file_signature_download,
    permission_document_file_signature_upload,
    permission_document_file_signature_verify,
    permission_document_file_signature_view
)
from .tasks import (
    task_refresh_signature_information,
    task_verify_missing_embedded_signature
)

logger = logging.getLogger(name=__name__)


class DocumentFileDetachedSignatureCreateView(
    ExternalObjectViewMixin, FormView
):
    external_object_permission = permission_document_file_sign_detached
    external_object_pk_url_kwarg = 'document_file_id'
    external_object_queryset = DocumentFile.valid.all()
    form_class = DocumentFileSignatureCreateForm
    view_icon = icon_document_file_signature_detached_create

    def form_valid(self, form):
        key = form.cleaned_data['key']
        passphrase = form.cleaned_data['passphrase'] or None

        try:
            DetachedSignature.objects.sign_document_file(
                document_file=self.external_object, key=key,
                passphrase=passphrase, user=self.request.user
            )
        except NeedPassphrase:
            messages.error(
                message=_(message='Passphrase is needed to unlock this key.'),
                request=self.request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(
                    kwargs={
                        'document_file_id': self.external_object.pk
                    },
                    viewname='signatures:document_file_signature_detached_create'
                )
            )
        except PassphraseError:
            messages.error(
                message=_(message='Passphrase is incorrect.'),
                request=self.request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(
                    kwargs={
                        'document_file_id': self.external_object.pk
                    },
                    viewname='signatures:document_file_signature_detached_create'
                )
            )
        else:
            messages.success(
                message=_(message='Document file signed successfully.'),
                request=self.request
            )

        return super().form_valid(form)

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                'Sign document file "%s" with a detached signature'
            ) % self.external_object
        }

    def get_form_extra_kwargs(self):
        return {'user': self.request.user}

    def get_post_action_redirect(self):
        return reverse(
            kwargs={'document_file_id': self.external_object.pk},
            viewname='signatures:document_file_signature_list'
        )


class DocumentFileEmbeddedSignatureCreateView(
    ExternalObjectViewMixin, FormView
):
    external_object_permission = permission_document_file_sign_embedded
    external_object_pk_url_kwarg = 'document_file_id'
    external_object_queryset = DocumentFile.valid.all()
    form_class = DocumentFileSignatureCreateForm
    view_icon = icon_document_file_signature_embedded_create

    def form_valid(self, form):
        key = form.cleaned_data['key']
        passphrase = form.cleaned_data['passphrase'] or None

        try:
            signature = EmbeddedSignature.objects.sign_document_file(
                document_file=self.external_object,
                key=key, passphrase=passphrase, user=self.request.user
            )
        except NeedPassphrase:
            messages.error(
                message=_(message='Passphrase is needed to unlock this key.'),
                request=self.request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(
                    kwargs={
                        'document_file_id': self.external_object.pk
                    },
                    viewname='signatures:document_file_signature_embedded_create'
                )
            )
        except PassphraseError:
            messages.error(
                message=_(message='Passphrase is incorrect.'),
                request=self.request
            )
            return HttpResponseRedirect(
                redirect_to=reverse(
                    kwargs={
                        'document_file_id': self.external_object.pk
                    },
                    viewname='signatures:document_file_signature_embedded_create'
                )
            )
        else:
            messages.success(
                message=_(message='Document file signed successfully.'),
                request=self.request
            )

            return HttpResponseRedirect(
                redirect_to=reverse(
                    kwargs={'document_file_id': signature.document_file.pk},
                    viewname='signatures:document_file_signature_list'
                )
            )

        return super().form_valid(form)

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                'Sign document file "%s" with a embedded signature'
            ) % self.external_object
        }

    def get_form_extra_kwargs(self):
        return {'user': self.request.user}


class DocumentFileDetachedSignatureDeleteView(SingleObjectDeleteView):
    object_permission = permission_document_file_signature_delete
    pk_url_kwarg = 'signature_id'
    view_icon = icon_document_file_signature_detached_delete

    def get_extra_context(self):
        return {
            'object': self.object.document_file,
            'signature': self.object,
            'title': _(message='Delete detached signature: %s') % self.object
        }

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'document_file_id': self.object.document_file.pk
            },
            viewname='signatures:document_file_signature_list'
        )

    def get_source_queryset(self):
        queryset_document_files = DocumentFile.valid.all()

        return DetachedSignature.objects.filter(
            document_file_id__in=queryset_document_files.values('pk')
        )


class DocumentFileDetachedSignatureDownloadView(SingleObjectDownloadView):
    object_permission = permission_document_file_signature_download
    pk_url_kwarg = 'signature_id'
    view_icon = icon_document_file_signature_detached_download

    def get_download_file_object(self):
        return self.object.signature_file

    def get_download_filename(self):
        return force_str(s=self.object)

    def get_source_queryset(self):
        queryset_document_files = DocumentFile.valid.all()

        return DetachedSignature.objects.filter(
            document_file_id__in=queryset_document_files.values('pk')
        )


class DocumentFileDetachedSignatureUploadView(
    ExternalObjectViewMixin, SingleObjectCreateView
):
    external_object_permission = permission_document_file_signature_upload
    external_object_pk_url_kwarg = 'document_file_id'
    external_object_queryset = DocumentFile.valid.all()
    fields = ('signature_file',)
    model = DetachedSignature
    view_icon = icon_document_file_signature_detached_upload

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                'Upload detached signature for document file: %s'
            ) % self.external_object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'document_file': self.external_object
        }

    def get_post_action_redirect(self):
        return reverse(
            kwargs={
                'document_file_id': self.external_object.pk
            }, viewname='signatures:document_file_signature_list'
        )


class DocumentFileSignatureDetailView(SingleObjectDetailView):
    form_class = DocumentFileSignatureDetailForm
    object_permission = permission_document_file_signature_view
    pk_url_kwarg = 'signature_id'
    view_icon = icon_document_file_signature_detail

    def get_extra_context(self):
        return {
            'hide_object': True,
            'object': self.object.document_file,
            'signature': self.object,
            'title': _(
                'Details for signature: %s'
            ) % self.object
        }

    def get_source_queryset(self):
        queryset_document_files = DocumentFile.valid.all()

        return SignatureBaseModel.objects.select_subclasses().filter(
            document_file_id__in=queryset_document_files.values('pk')
        )


class DocumentFileSignatureListView(
    ExternalObjectViewMixin, SingleObjectListView
):
    external_object_permission = permission_document_file_signature_view
    external_object_pk_url_kwarg = 'document_file_id'
    external_object_queryset = DocumentFile.valid.all()
    view_icon = icon_document_file_signature_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_document_file_signature_list,
            'no_results_text': _(
                'Signatures help provide authorship evidence and tamper '
                'detection. They are very secure and hard to '
                'forge. A signature can be embedded as part of the document '
                'itself or uploaded as a separate file.'
            ),
            'no_results_secondary_links': [
                link_document_file_signature_detached_create.resolve(
                    RequestContext(
                        dict_={
                            'object': self.external_object
                        }, request=self.request
                    )
                ),
                link_document_file_signature_embedded_create.resolve(
                    RequestContext(
                        dict_={
                            'object': self.external_object
                        }, request=self.request
                    )
                ),
                link_document_file_signature_detached_upload.resolve(
                    RequestContext(
                        dict_={
                            'object': self.external_object
                        }, request=self.request
                    )
                )
            ],
            'no_results_title': _(
                'There are no signatures for this document file.'
            ),
            'object': self.external_object,
            'title': _(
                'Signatures for document file: %s'
            ) % self.external_object
        }

    def get_source_queryset(self):
        return self.external_object.signatures.all()


class AllDocumentSignatureRefreshView(ConfirmView):
    extra_context = {
        'message': _(
            'On large databases this operation may take some time '
            'to execute.'
        ), 'title': _(message='Refresh all signatures information?')
    }
    view_icon = icon_document_file_all_signature_refresh
    view_permission = permission_document_file_signature_verify

    def get_post_action_redirect(self):
        return reverse(viewname='common:tools_list')

    def view_action(self):
        task_refresh_signature_information.apply_async()
        messages.success(
            message=_(
                message='Signature information refresh queued successfully.'
            ), request=self.request
        )


class AllDocumentSignatureVerifyView(ConfirmView):
    extra_context = {
        'message': _(
            'On large databases this operation may take some time to '
            'execute.'
        ), 'title': _(message='Verify all document for signatures?')
    }
    view_icon = icon_document_file_all_signature_verify
    view_permission = permission_document_file_signature_verify

    def get_post_action_redirect(self):
        return reverse(viewname='common:tools_list')

    def view_action(self):
        task_verify_missing_embedded_signature.apply_async()
        messages.success(
            message=_(message='Signature verification queued successfully.'),
            request=self.request
        )



from django.utils.translation import gettext_lazy as _

from .classes import Layer
from .icons import icon_layer_decorations, icon_layer_transformation
from .permissions import (
    permission_transformation_create, permission_transformation_delete,
    permission_transformation_edit, permission_transformation_view
)

layer_decorations = Layer(
    empty_results_text=_(
        message='Decorations are visual elements that add contextual '
        'information to images.'
    ), icon=icon_layer_decorations, label=_(message='Decorations'),
    name='decorations', order=10, permission_map={
        'create': permission_transformation_create,
        'delete': permission_transformation_delete,
        'edit': permission_transformation_edit,
        'select': permission_transformation_create,
        'view': permission_transformation_view
    }
)

layer_saved_transformations = Layer(
    default=True, icon=icon_layer_transformation, label=_(
        message='Saved transformations'
    ), name='saved_transformations', order=100, permission_map={
        'create': permission_transformation_create,
        'delete': permission_transformation_delete,
        'edit': permission_transformation_edit,
        'select': permission_transformation_create,
        'view': permission_transformation_view
    }
)



from django.urls import re_path

from .views import (
    StatisticNamespaceDetailView, StatisticNamespaceListView,
    StatisticTypeDetailView, StatisticTypeQueueView
)

urlpatterns = [
    re_path(
        route=r'^namespaces/$', view=StatisticNamespaceListView.as_view(),
        name='statistic_namespace_list'
    ),
    re_path(
        route=r'^namespaces/(?P<slug>[\w-]+)/$',
        view=StatisticNamespaceDetailView.as_view(),
        name='statistic_namespace_detail'
    ),
    re_path(
        route=r'^namespaces/statistics/(?P<slug>[\w-]+)/view/$',
        view=StatisticTypeDetailView.as_view(), name='statistic_detail'
    ),
    re_path(
        route=r'^namespaces/statistics/(?P<slug>[\w-]+)/queue/$',
        view=StatisticTypeQueueView.as_view(), name='statistic_queue'
    )
]



from django.utils.translation import gettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_(message='Credentials'), name='credentials'
)

event_credential_created = namespace.add_event_type(
    label=_(message='Credential created'), name='credential_created'
)
event_credential_edited = namespace.add_event_type(
    label=_(message='Credential edited'), name='credential_edited'
)
event_credential_used = namespace.add_event_type(
    label=_(message='Credential used'), name='credential_used'
)



from django.shortcuts import get_object_or_404

from mayan.apps.documents.models.document_models import Document
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from ..permissions import (
    permission_workflow_instance_transition,
    permission_workflow_template_view, permission_workflow_tools
)
from ..serializers.workflow_instance_serializers import (
    WorkflowInstanceLaunchSerializer, WorkflowInstanceLogEntrySerializer,
    WorkflowInstanceSerializer
)
from ..serializers.workflow_template_transition_serializers import WorkflowTemplateTransitionSerializer


class APIWorkflowInstanceLaunchActionView(generics.ObjectActionAPIView):
    """
    post: Launch a new workflow instance for the specified document.
    """
    lookup_url_kwarg = 'document_id'
    mayan_object_permission_map = {'POST': permission_workflow_tools}
    serializer_class = WorkflowInstanceLaunchSerializer
    source_queryset = Document.valid.all()

    def get_serializer_extra_context(self):
        obj = self.get_object()
        return {
            'document': obj, 'document_type': obj.document_type
        }

    def object_action(self, obj, request, serializer):
        workflow_template = serializer.validated_data['workflow_template_id']
        workflow_template.launch_for(document=obj)


class APIWorkflowInstanceListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the document workflow instances.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view
    }
    mayan_object_permission_map = {'GET': permission_workflow_template_view}
    serializer_class = WorkflowInstanceSerializer

    def get_source_queryset(self):
        return self.get_external_object().workflows.all()


class APIWorkflowInstanceDetailView(
    ExternalObjectAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Return the details of the selected document workflow instances.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    lookup_url_kwarg = 'workflow_instance_id'
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view
    }
    mayan_object_permission_map = {'GET': permission_workflow_template_view}
    serializer_class = WorkflowInstanceSerializer

    def get_source_queryset(self):
        return self.get_external_object().workflows.all()


class APIWorkflowInstanceLogEntryDetailView(
    ExternalObjectAPIViewMixin, generics.RetrieveAPIView
):
    """
    get: Return the details of the selected document instances log entry.
    """
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view
    }
    serializer_class = WorkflowInstanceLogEntrySerializer
    lookup_url_kwarg = 'workflow_instance_log_entry_id'

    def get_source_queryset(self):
        return self.get_workflow_instance().log_entries.all()

    def get_workflow_instance(self):
        workflow = get_object_or_404(
            klass=self.get_external_object().workflows,
            pk=self.kwargs['workflow_instance_id']
        )

        return workflow


class APIWorkflowInstanceLogEntryListView(
    ExternalObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of all the document workflow instances log entries.
    post: Transition a document workflow by creating a new document workflow instance log entry.
    """
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view,
        'POST': permission_workflow_instance_transition
    }
    mayan_object_permission_map = {'GET': permission_workflow_template_view}
    ordering_fields = (
        'comment', 'id', 'transition', 'transition__destination_state',
        'transition__origin_state'
    )
    serializer_class = WorkflowInstanceLogEntrySerializer

    def get_serializer_extra_context(self):
        if self.kwargs:
            return {
                'workflow_instance': self.get_workflow_instance()
            }
        else:
            return {}

    def get_source_queryset(self):
        return self.get_workflow_instance().log_entries.all()

    def get_workflow_instance(self):
        workflow = get_object_or_404(
            klass=self.get_external_object().workflows,
            pk=self.kwargs['workflow_instance_id']
        )

        return workflow


class APIWorkflowInstanceLogEntryTransitionListView(
    ExternalObjectAPIViewMixin, generics.ListAPIView
):
    """
    get: Returns a list of all the possible transition choices for the workflow instance.
    """
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    mayan_external_object_permission_map = {
        'GET': permission_workflow_template_view
    }
    mayan_object_permission_map = {'GET': permission_workflow_template_view}
    ordering_fields = ('destination_state', 'id', 'origin_state')
    serializer_class = WorkflowTemplateTransitionSerializer

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.kwargs:
            context.update(
                {
                    'workflow_instance': self.get_workflow_instance()
                }
            )

        return context

    def get_source_queryset(self):
        return self.get_workflow_instance().get_transition_choices(
            user=self.request.user
        )

    def get_workflow_instance(self):
        workflow = get_object_or_404(
            klass=self.get_external_object().workflows,
            pk=self.kwargs['workflow_instance_id']
        )

        return workflow



from mayan.apps.documents.permissions import (
    permission_document_version_view, permission_document_view
)
from mayan.apps.documents.search import (
    search_model_document, search_model_document_version,
    search_model_document_version_page
)
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.dynamic_search.tests.mixins.base import SearchTestMixin

from .mixins import DocumentVersionOCRTestMixin


class DocumentOCRSearchTestCase(
    DocumentVersionOCRTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    _test_search_model = search_model_document
    auto_create_test_document_version_ocr_content = True

    def test_search_model_document_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'versions__version_pages__ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_document_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'versions__version_pages__ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_search_model_document_with_access(self):
        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'versions__version_pages__ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentVersionOCRSearchTestCase(
    DocumentVersionOCRTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    _test_search_model = search_model_document_version
    auto_create_test_document_version_ocr_content = True

    def test_search_model_document_version_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'version_pages__ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_document_version_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'version_pages__ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document_version in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_search_model_document_version_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'version_pages__ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document_version not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class DocumentVersionPageOCRSearchTestCase(
    DocumentVersionOCRTestMixin, SearchTestMixin, GenericDocumentViewTestCase
):
    _test_search_model = search_model_document_version_page
    auto_create_test_document_version_ocr_content = True

    def test_search_model_document_version_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_document_version_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document_version_page in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_search_model_document_version_with_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_version_view
        )

        self._test_document.delete()

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'ocr_content__content': self._test_document_version_page.ocr_content.content
            }
        )
        self.assertTrue(self._test_document_version_page not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.documents.tests.mixins.document_version_mixins import DocumentVersionTestMixin

from ..events import event_ocr_document_version_finished
from ..models import DocumentVersionPageOCRContent
from ..tasks import task_document_version_ocr_process

from .literals import (
    TEST_DOCUMENT_VERSION_OCR_CONTENT,
    TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
)


class DocumentTypeOCRSettingsAPIViewTestMixin:
    def _request_test_document_type_ocr_settings_details_api_view(self):
        return self.get(
            viewname='rest_api:document-type-ocr-settings-view',
            kwargs={'document_type_id': self._test_document_type.pk}
        )

    def _request_test_document_type_ocr_settings_patch_api_view(self):
        return self.patch(
            viewname='rest_api:document-type-ocr-settings-view',
            kwargs={'document_type_id': self._test_document_type.pk},
            data={'auto_ocr': True}
        )

    def _request_test_document_type_ocr_settings_put_api_view(self):
        return self.put(
            viewname='rest_api:document-type-ocr-settings-view',
            kwargs={'document_type_id': self._test_document_type.pk},
            data={'auto_ocr': True}
        )


class DocumentTypeOCRViewTestMixin:
    def _request_test_document_type_ocr_settings_view(self):
        return self.get(
            viewname='ocr:document_type_ocr_settings', kwargs={
                'document_type_id': self._test_document_type.pk
            }
        )

    def _request_document_type_ocr_submit_view(self):
        return self.post(
            viewname='ocr:document_type_submit', data={
                'document_type': [self._test_document_type.pk]
            }
        )


class DocumentOCRAPIViewTestMixin:
    def _request_test_document_ocr_submit_api_view(self):
        return self.post(
            viewname='rest_api:document-ocr-submit-view',
            kwargs={'document_id': self._test_document.pk}
        )


class DocumentVersionOCRTaskTestMixin:
    def _execute_task_document_version_ocr_process(self):
        task_document_version_ocr_process.apply_async(
            kwargs={
                'document_version_id': self._test_document_version.pk,
                'user_id': self._test_case_user.pk
            }
        ).get()


class DocumentVersionOCRTestMixin(DocumentVersionTestMixin):
    auto_create_test_document_version_ocr_content = False

    def setUp(self):
        super().setUp()

        if self.auto_create_test_document_version_ocr_content:
            self._create_test_document_version_ocr_content()

    def _create_test_document_version_ocr_content(self):
        DocumentVersionPageOCRContent.objects.create(
            document_version_page=self._test_document_version_page,
            content=TEST_DOCUMENT_VERSION_OCR_CONTENT
        )
        event_ocr_document_version_finished.commit(
            action_object=self._test_document,
            target=self._test_document_version
        )

    def _do_test_document_version_ocr_content_delete(self):
        DocumentVersionPageOCRContent.objects.delete_content_for(
            document_version=self._test_document_version
        )


class DocumentVersionOCRAPIViewTestMixin(DocumentVersionOCRTestMixin):
    def _request_test_document_version_ocr_submit_api_view(self):
        return self.post(
            viewname='rest_api:document-version-ocr-submit-view', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk
            }
        )


class DocumentVersionOCRViewTestMixin(DocumentVersionOCRTestMixin):
    def _request_test_document_version_ocr_content_detail_view(self):
        return self.get(
            viewname='ocr:document_version_ocr_content_view', kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_content_single_delete_view(self):
        return self.post(
            viewname='ocr:document_version_ocr_content_single_delete',
            kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_content_multiple_delete_view(self):
        return self.post(
            viewname='ocr:document_version_ocr_content_multiple_delete',
            data={
                'id_list': self._test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_single_submit_view(self):
        return self.post(
            viewname='ocr:document_version_ocr_single_submit', kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_multiple_submit_view(self):
        return self.post(
            viewname='ocr:document_version_ocr_multiple_submit', data={
                'id_list': self._test_document_version.pk
            }
        )

    def _request_test_document_version_ocr_content_download_view(self):
        return self.get(
            viewname='ocr:document_version_ocr_content_download', kwargs={
                'document_version_id': self._test_document_version.pk
            }
        )


class DocumentVersionPageOCRAPIViewTestMixin:
    def _request_test_document_version_page_ocr_content_detail_api_view_via_get(self):
        return self.get(
            viewname='rest_api:document-version-page-ocr-content-detail-view', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
                'document_version_page_id': self._test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_ocr_content_edit_api_view_via_patch(self):
        return self.patch(
            viewname='rest_api:document-version-page-ocr-content-detail-view', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
                'document_version_page_id': self._test_document_version_page.pk
            }, data={
                'content': TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
            }
        )

    def _request_test_document_version_page_ocr_content_edit_api_view_via_put(self):
        return self.put(
            viewname='rest_api:document-version-page-ocr-content-detail-view', kwargs={
                'document_id': self._test_document.pk,
                'document_version_id': self._test_document_version.pk,
                'document_version_page_id': self._test_document_version_page.pk
            }, data={
                'content': TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
            }
        )


class DocumentVersionPageOCRViewTestMixin:
    def _request_test_document_version_page_ocr_content_detail_view(self):
        return self.get(
            viewname='ocr:document_version_page_ocr_content_detail_view', kwargs={
                'document_version_page_id': self._test_document_version_page.pk
            }
        )

    def _request_test_document_version_page_ocr_content_edit_view(self):
        return self.post(
            viewname='ocr:document_version_page_ocr_content_edit_view', kwargs={
                'document_version_page_id': self._test_document_version_page.pk
            }, data={
                'content': TEST_DOCUMENT_VERSION_PAGE_OCR_CONTENT_UPDATED
            }
        )



from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL)
    ]

    operations = [
        migrations.CreateModel(
            name='UserOTPData',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                (
                    'secret', models.CharField(
                        blank=True, max_length=96, verbose_name='16 '
                        'character, base32 encoded random string.'
                    )
                ),
                (
                    'user', models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='otp_data',
                        to=settings.AUTH_USER_MODEL, verbose_name='User'
                    )
                )
            ],
            options={
                'verbose_name': 'User OTP data',
                'verbose_name_plural': 'Users OTP data'
            }
        )
    ]



from unittest import mock

from mayan.apps.documents.tests.base import GenericDocumentTestCase

from ..events import event_ocr_document_version_finished
from ..exceptions import OCRError

from .mixins import DocumentVersionOCRTaskTestMixin


class OCRTaskTestCase(
    DocumentVersionOCRTaskTestMixin, GenericDocumentTestCase
):
    @mock.patch(
        'mayan.apps.ocr.backends.tesseract.Tesseract._execute'
    )
    def test_task_document_version_ocr_finished_error_create(
        self, method_backend_execute
    ):
        TEST_EXCEPTION_TEXT = 'Fake exception'

        method_backend_execute.side_effect = OCRError(
            TEST_EXCEPTION_TEXT
        )

        self._clear_events()

        self._execute_task_document_version_ocr_process()

        self.assertEqual(
            self._test_document_version_page.error_log.count(), 1
        )
        self.assertEqual(
            self._test_document_version_page.error_log.first().text,
            TEST_EXCEPTION_TEXT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_document)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_version)
        self.assertEqual(
            events[0].verb, event_ocr_document_version_finished.id
        )

    @mock.patch(
        'mayan.apps.ocr.backends.tesseract.Tesseract._execute'
    )
    def test_task_document_version_ocr_finished_unhandled_error(
        self, method_backend_execute
    ):
        TEST_EXCEPTION_TEXT = 'Fake exception'

        self._silence_logger(name='mayan.apps.ocr.managers')

        class TestException(Exception):
            """Dummy unexpected OCR exception for testing."""

        method_backend_execute.side_effect = TestException(
            TEST_EXCEPTION_TEXT
        )

        self._clear_events()

        with self.assertRaises(expected_exception=TestException):
            self._execute_task_document_version_ocr_process()

        self.assertEqual(
            self._test_document_version_page.error_log.count(), 0
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.urls import re_path

from .api_views import (
    APICurrentUserView, APIGroupDetailView, APIGroupListView,
    APIGroupUserAddView, APIGroupUserListView, APIGroupUserRemoveView,
    APIUserDetailView, APIUserGroupListView, APIUserListView
)
from .views.group_views import (
    GroupCreateView, GroupDeleteView, GroupDetailView, GroupEditView,
    GroupListView, GroupUserAddRemoveView
)
from .views.user_views import (
    UserCreateView, UserDeleteView, UserDetailView, UserEditView,
    UserGroupAddRemoveView, UserListView, UserOptionsEditView
)

urlpatterns_groups = [
    re_path(
        route=r'^groups/$', name='group_list', view=GroupListView.as_view()
    ),
    re_path(
        route=r'^groups/create/$', name='group_create',
        view=GroupCreateView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<group_id>\d+)/delete/$',
        name='group_single_delete', view=GroupDeleteView.as_view()
    ),
    re_path(
        route=r'^groups/multiple/delete/$', name='group_multiple_delete',
        view=GroupDeleteView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<group_id>\d+)/$', name='group_detail',
        view=GroupDetailView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<group_id>\d+)/edit/$', name='group_edit',
        view=GroupEditView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<group_id>\d+)/users/$', name='group_members',
        view=GroupUserAddRemoveView.as_view()
    )
]

urlpatterns_users = [
    re_path(
        route=r'^users/$', name='user_list', view=UserListView.as_view()
    ),
    re_path(
        route=r'^users/create/$', name='user_create',
        view=UserCreateView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>\d+)/delete/$', name='user_single_delete',
        view=UserDeleteView.as_view()
    ),
    re_path(
        route=r'^users/multiple/delete/$', name='user_multiple_delete',
        view=UserDeleteView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>\d+)/$', name='user_details',
        view=UserDetailView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>\d+)/edit/$', name='user_edit',
        view=UserEditView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>\d+)/groups/$', name='user_groups',
        view=UserGroupAddRemoveView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>\d+)/options/$', name='user_options',
        view=UserOptionsEditView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_groups)
urlpatterns.extend(urlpatterns_users)

api_urls = [
    re_path(
        route=r'^groups/$', name='group-list',
        view=APIGroupListView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<group_id>[0-9]+)/$', name='group-detail',
        view=APIGroupDetailView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<group_id>[0-9]+)/users/$',
        name='group-user-list', view=APIGroupUserListView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<group_id>[0-9]+)/users/add/$',
        name='group-user-add', view=APIGroupUserAddView.as_view()
    ),
    re_path(
        route=r'^groups/(?P<group_id>[0-9]+)/users/remove/$',
        name='group-user-remove', view=APIGroupUserRemoveView.as_view()
    ),
    re_path(
        route=r'^users/$', name='user-list', view=APIUserListView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>[0-9]+)/$', name='user-detail',
        view=APIUserDetailView.as_view()
    ),
    re_path(
        route=r'^users/current/$', name='user-current',
        view=APICurrentUserView.as_view()
    ),
    re_path(
        route=r'^users/(?P<user_id>[0-9]+)/groups/$', name='user-group-list',
        view=APIUserGroupListView.as_view()
    )
]



DEFAULT_AUTHENTICATION_OIDC_USER_PROFILE_URL = None

DEFAULT_OIDC_OP_AUTHORIZATION_ENDPOINT = None
DEFAULT_OIDC_OP_JWKS_ENDPOINT = None
DEFAULT_OIDC_OP_TOKEN_ENDPOINT = None
DEFAULT_OIDC_OP_USER_ENDPOINT = None
DEFAULT_OIDC_RP_CLIENT_ID = None
DEFAULT_OIDC_RP_CLIENT_SECRET = None
DEFAULT_OIDC_RP_SIGN_ALGO = 'HS256'

DEFAULT_OIDC_USERNAME_ALGO = 'mayan.apps.authentication_oidc.utils.generate_username_from_email'



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0007_auto_20150711_0656')
    ]

    operations = [
        migrations.AlterField(
            model_name='transformation',
            name='order',
            field=models.PositiveIntegerField(
                default=0, help_text='Order in which the transformations '
                'will be executed.', db_index=True, verbose_name='Order',
                blank=True
            ),
            preserve_default=True
        ),
        migrations.AlterUniqueTogether(
            name='transformation',
            unique_together={
                ('content_type', 'object_id', 'order')
            }
        )
    ]



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('permissions', '0002_auto_20150628_0533'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='role',
            name='name',
        ),
    ]



from django.apps import apps


def purge_periodic_tasks():
    IntervalSchedule = apps.get_model(
        app_label='django_celery_beat', model_name='IntervalSchedule'
    )
    PeriodicTask = apps.get_model(
        app_label='django_celery_beat', model_name='PeriodicTask'
    )

    PeriodicTask.objects.all().delete()
    IntervalSchedule.objects.all().delete()



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db.models.signals import post_save, pre_delete

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.menus import menu_list_facet
from mayan.apps.events.classes import EventModelRegistry, ModelEventType

from .events import event_error_log_deleted
from .links import link_object_error_log_entry_list
from .literals import DEFAULT_ERROR_LOG_PARTITION_ENTRY_LIMIT
from .permissions import (
    permission_error_log_entry_delete, permission_error_log_entry_view
)


class ErrorLog:
    _registry = {}

    @staticmethod
    def get_model_instance_partition_name(model_instance):
        return '{}.{}'.format(model_instance._meta.label, model_instance.pk)

    @classmethod
    def all(cls):
        return cls._registry

    @classmethod
    def get(cls, name):
        return cls._registry[name]

    def __init__(
        self, app_config, limit=DEFAULT_ERROR_LOG_PARTITION_ENTRY_LIMIT
    ):
        self.app_config = app_config
        self.limit = limit

        self.__class__._registry[app_config.name] = self

    def __str__(self):
        return str(self.app_config.verbose_name)

    def register_model(self, model, register_permission=True):
        error_log_instance = self

        if getattr(model, 'error_log', None):
            raise ImproperlyConfigured(
                'Model `{}` has already been registered for error '
                'logging.'.format(model)
            )

        @property
        def method_instance_logs(self):
            ContentType = apps.get_model(
                app_label='contenttypes', model_name='ContentType'
            )

            content_type = ContentType.objects.get_for_model(
                model=self
            )

            error_log_partition, created = error_log_instance.stored_error_log.partitions.get_or_create(
                content_type=content_type, object_id=self.pk, defaults={
                    'name': ErrorLog.get_model_instance_partition_name(
                        model_instance=self
                    )
                }
            )

            error_log_partition.entries.exclude(
                pk__in=error_log_partition.entries.order_by('-datetime')[:error_log_instance.limit].values('pk')
            ).delete()

            return error_log_partition.entries

        EventModelRegistry.register(model=model)

        ModelEventType.register(
            event_types=(event_error_log_deleted,), model=model
        )

        model.add_to_class(name='error_log', value=method_instance_logs)

        menu_list_facet.bind_links(
            links=(link_object_error_log_entry_list,), sources=(model,)
        )

        if register_permission:
            ModelPermission.register(
                model=model, permissions=(
                    permission_error_log_entry_delete,
                    permission_error_log_entry_view
                )
            )

        def handler_model_instance_delete_partition(
            sender, instance, **kwargs
        ):
            ContentType = apps.get_model(
                app_label='contenttypes', model_name='ContentType'
            )

            content_type = ContentType.objects.get_for_model(
                model=instance
            )

            return self.stored_error_log.partitions.filter(
                content_type=content_type, object_id=instance.pk
            ).delete()

        def handler_model_instance_create_partition(
            sender, instance, **kwargs
        ):
            if kwargs['created']:
                ContentType = apps.get_model(
                    app_label='contenttypes', model_name='ContentType'
                )

                content_type = ContentType.objects.get_for_model(
                    model=instance
                )

                return self.stored_error_log.partitions.create(
                    content_type=content_type, object_id=instance.pk,
                    name=ErrorLog.get_model_instance_partition_name(
                        model_instance=instance
                    )
                )

        post_save.connect(
            dispatch_uid='logging_handler_model_instance_create_partition',
            receiver=handler_model_instance_create_partition,
            sender=model, weak=False
        )
        pre_delete.connect(
            dispatch_uid='logging_handler_model_instance_delete_partition',
            receiver=handler_model_instance_delete_partition,
            sender=model, weak=False
        )

    @property
    def stored_error_log(self):
        StoredErrorLog = apps.get_model(
            app_label='logging', model_name='StoredErrorLog'
        )
        try:
            stored_error_log, created = StoredErrorLog.objects.get_or_create(
                name=self.app_config.name
            )
        except StoredErrorLog.MultipleObjectsReturned:
            # Self heal previously repeated entries.
            StoredErrorLog.objects.filter(name=self.app_config.name).delete()
            stored_error_log, created = StoredErrorLog.objects.get_or_create(
                name=self.app_config.name
            )

        return stored_error_log



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_LOGGING_DISABLE_COLOR_FORMATTER, DEFAULT_LOGGING_ENABLE,
    DEFAULT_LOGGING_HANDLERS, DEFAULT_LOGGING_LEVEL,
    DEFAULT_LOGGING_LOG_FILE_PATH, LOGGING_HANDLER_OPTIONS
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Logging'), name='logging'
)

setting_logging_disable_color_formatter = setting_namespace.do_setting_add(
    default=DEFAULT_LOGGING_DISABLE_COLOR_FORMATTER,
    global_name='LOGGING_DISABLE_COLOR_FORMATTER', help_text=_(
        'Disable logging message ANSI color codes.'
    )
)
setting_logging_enable = setting_namespace.do_setting_add(
    choices=('false', 'true'), default=DEFAULT_LOGGING_ENABLE,
    global_name='LOGGING_ENABLE', help_text=_(
        'Automatically enable logging to all apps.'
    )
)
setting_logging_handlers = setting_namespace.do_setting_add(
    choices=LOGGING_HANDLER_OPTIONS, default=DEFAULT_LOGGING_HANDLERS,
    global_name='LOGGING_HANDLERS', help_text=_(
        'List of handlers to which logging messages will be sent.'
    )
)
setting_logging_level = setting_namespace.do_setting_add(
    default=DEFAULT_LOGGING_LEVEL, global_name='LOGGING_LEVEL', help_text=_(
        'Level for the logging system.'
    )
)
setting_logging_log_file_path = setting_namespace.do_setting_add(
    default=DEFAULT_LOGGING_LOG_FILE_PATH,
    global_name='LOGGING_LOG_FILE_PATH', help_text=_(
        'Path to the logfile that will track errors during production.'
    ), is_path=True
)



from django.urls import re_path

from .api_views import APIDocumentFileDownloadView
from .views import DocumentDownloadView, DocumentFileDownloadView

urlpatterns = [
    re_path(
        route=r'^documents/multiple/download/$',
        name='document_download_multiple',
        view=DocumentDownloadView.as_view()
    ),
    re_path(
        route=r'^documents/(?P<document_id>\d+)/download/$',
        name='document_download_single', view=DocumentDownloadView.as_view()
    ),
    re_path(
        route=r'^documents/files/(?P<document_file_id>\d+)/download/$',
        name='document_file_download',
        view=DocumentFileDownloadView.as_view()
    )
]

api_urls = [
    re_path(
        route=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/download/$',
        name='documentfile-download',
        view=APIDocumentFileDownloadView.as_view()
    )
]



from django import forms
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _, gettext

from mayan.apps.views.widgets import TextAreaDiv

from .models import DocumentVersionPageOCRContent


class DocumentVersionPageOCRContentDetailForm(forms.Form):
    contents = forms.CharField(
        label=_(message='Contents'),
        widget=TextAreaDiv(
            attrs={
                'class': 'full-height',
                'data-height-difference': 360
            }
        )
    )

    def __init__(self, *args, **kwargs):
        page = kwargs.pop('instance', None)
        super().__init__(*args, **kwargs)
        content = ''
        self.fields['contents'].initial = ''

        try:
            page_content = page.ocr_content.content
        except DocumentVersionPageOCRContent.DoesNotExist:
            """
            Not critical, just ignore and display empty content.
            """
        else:
            content = conditional_escape(
                str(page_content)
            )

        self.fields['contents'].initial = mark_safe(s=content)


class DocumentVersionPageOCRContentEditForm(forms.ModelForm):
    content = forms.CharField(
        label=_(message='Contents'),
        widget=forms.widgets.Textarea(
            attrs={
                'class': 'full-height',
                'data-height-difference': 360
            }
        )
    )

    class Meta:
        fields = ('content',)
        model = DocumentVersionPageOCRContent


class DocumentVersionOCRContentForm(forms.Form):
    """
    Form that concatenates all of a document pages' text content into a
    single textarea widget
    """
    contents = forms.CharField(
        label=_(message='Contents'),
        widget=TextAreaDiv(
            attrs={
                'class': 'full-height',
                'data-height-difference': 360
            }
        )
    )

    def __init__(self, *args, **kwargs):
        self.document = kwargs.pop('instance', None)
        super().__init__(*args, **kwargs)
        content = []
        self.fields['contents'].initial = ''
        try:
            document_pages = self.document.pages.all()
        except AttributeError:
            document_pages = []

        for page in document_pages:
            try:
                page_content = page.ocr_content.content
            except DocumentVersionPageOCRContent.DoesNotExist:
                """
                Not critical, just ignore and proceed to next page.
                """
            else:
                content.append(
                    conditional_escape(
                        text=str(page_content)
                    )
                )
                content.append(
                    '\n\n\n<hr/><div class="document-page-content-divider">- %s -</div><hr/>\n\n\n' % (
                        gettext(
                            'Page %(page_number)d'
                        ) % {'page_number': page.page_number}
                    )
                )

        self.fields['contents'].initial = mark_safe(
            s=''.join(content)
        )



from django.template import RequestContext
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _

from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.views.generics import (
    MultipleObjectDeleteView, RelationshipView, SingleObjectCreateView,
    SingleObjectEditView, SingleObjectListView
)

from ..forms import (
    DocumentTypeMetadataTypeRelationshipFormSet, MetadataTypeForm
)
from ..icons import (
    icon_document_type_metadata_type_list, icon_metadata,
    icon_metadata_type_create, icon_metadata_type_document_type_list,
    icon_metadata_type_edit, icon_metadata_type_list,
    icon_metadata_type_single_delete
)
from ..links import link_metadata_type_create
from ..models import MetadataType
from ..permissions import (
    permission_metadata_type_create, permission_metadata_type_delete,
    permission_metadata_type_edit, permission_metadata_type_view
)


class MetadataTypeCreateView(SingleObjectCreateView):
    extra_context = {
        'title': _(message='Create metadata type')
    }
    form_class = MetadataTypeForm
    model = MetadataType
    post_action_redirect = reverse_lazy(
        viewname='metadata:metadata_type_list'
    )
    view_icon = icon_metadata_type_create
    view_permission = permission_metadata_type_create

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class MetadataTypeDeleteView(MultipleObjectDeleteView):
    error_message = _(
        'Error deleting metadata type "%(instance)s"; %(exception)s'
    )
    model = MetadataType
    object_permission = permission_metadata_type_delete
    pk_url_kwarg = 'metadata_type_id'
    post_action_redirect = reverse_lazy(
        viewname='metadata:metadata_type_list'
    )
    success_message_plural = _(
        '%(count)d metadata types deleted successfully.'
    )
    success_message_single = _(
        'Metadata type "%(object)s" deleted successfully.'
    )
    success_message_singular = _(
        '%(count)d metadata type deleted successfully.'
    )
    title_plural = _(message='Delete the %(count)d selected metadata types.')
    title_single = _(message='Delete metadata type: %(object)s.')
    title_singular = _(message='Delete the %(count)d selected metadata type.')
    view_icon = icon_metadata_type_single_delete


class MetadataTypeEditView(SingleObjectEditView):
    form_class = MetadataTypeForm
    model = MetadataType
    object_permission = permission_metadata_type_edit
    pk_url_kwarg = 'metadata_type_id'
    post_action_redirect = reverse_lazy(
        viewname='metadata:metadata_type_list'
    )
    view_icon = icon_metadata_type_edit

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(message='Edit metadata type: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class MetadataTypeListView(SingleObjectListView):
    model = MetadataType
    object_permission = permission_metadata_type_view
    view_icon = icon_metadata_type_list

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_metadata,
            'no_results_main_link': link_metadata_type_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Metadata types are user defined properties that can be '
                'assigned values. Once created they must be associated to '
                'document types, either as optional or required, for each. '
                'Setting a metadata type as required for a document type '
                'will block the upload of documents of that type until a '
                'metadata value is provided.'
            ),
            'no_results_title': _(message='There are no metadata types'),
            'title': _(message='Metadata types')
        }


class DocumentTypeMetadataTypeRelationshipView(RelationshipView):
    form_class = DocumentTypeMetadataTypeRelationshipFormSet
    model = DocumentType
    model_permission = permission_document_type_edit
    pk_url_kwarg = 'document_type_id'
    relationship_related_field = 'metadata'
    relationship_related_query_field = 'metadata_type'
    sub_model = MetadataType
    sub_model_permission = permission_metadata_type_edit
    view_icon = icon_document_type_metadata_type_list

    def get_extra_context(self):
        return {
            'form_display_mode_table': True,
            'no_results_icon': icon_metadata,
            'no_results_main_link': link_metadata_type_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Create metadata type relationships to be able to associate '
                'them to this document type.'
            ),
            'no_results_title': _(
                'There are no metadata type relationships available'
            ),
            'object': self.get_object(),
            'title': _(
                'Metadata type relationships for document type: %s'
            ) % self.get_object()
        }

    def get_form_extra_kwargs(self):
        return {
            '_event_actor': self.request.user
        }

    def get_initial(self):
        obj = self.get_object()
        initial = []

        for element in self.get_sub_model_queryset():
            initial.append(
                {
                    'object': obj,
                    'relationship_related_field': self.relationship_related_field,
                    'relationship_related_query_field': self.relationship_related_query_field,
                    'sub_object': element
                }
            )
        return initial

    def get_post_action_redirect(self):
        return reverse(viewname='documents:document_type_list')


class MetadataTypesDocumentTypeRelationshipView(
    DocumentTypeMetadataTypeRelationshipView
):
    model = MetadataType
    model_permission = permission_metadata_type_edit
    pk_url_kwarg = 'metadata_type_id'
    relationship_related_field = 'document_types'
    relationship_related_query_field = 'document_type'
    sub_model = DocumentType
    sub_model_permission = permission_document_type_edit
    view_icon = icon_metadata_type_document_type_list

    def get_extra_context(self):
        return {
            'form_display_mode_table': True,
            'object': self.get_object(),
            'title': _(
                'Document type relationships for metadata type: %s'
            ) % self.get_object()
        }

    def get_initial(self):
        obj = self.get_object()
        initial = []

        for element in self.get_sub_model_queryset():
            initial.append(
                {
                    'object': obj,
                    'relationship_related_field': self.relationship_related_field,
                    'relationship_related_query_field': self.relationship_related_query_field,
                    'sub_object': element
                }
            )
        return initial

    def get_post_action_redirect(self):
        return reverse(viewname='metadata:metadata_type_list')



from unittest import mock

from django_celery_beat.models import PeriodicTask

from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_version_edited, event_document_version_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.sources.events import (
    event_source_created, event_source_edited
)

from .mixins import PeriodicSourceBackendTestMixin


class PeriodicSourceBackendTestCase(
    PeriodicSourceBackendTestMixin, GenericDocumentTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def test_periodic_source_create(self):
        periodic_task_count = PeriodicTask.objects.count()

        self._clear_events()

        self._test_source_create()

        self.assertEqual(
            PeriodicTask.objects.count(), periodic_task_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_source)
        self.assertEqual(events[0].target, self._test_source)
        self.assertEqual(events[0].verb, event_source_created.id)

    def test_periodic_source_delete(self):
        self._test_source_create()

        periodic_task_count = PeriodicTask.objects.count()

        self._clear_events()

        self._test_source.delete()

        self.assertEqual(
            PeriodicTask.objects.count(), periodic_task_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_periodic_source_update(self):
        self._test_source_create()

        periodic_task_count = PeriodicTask.objects.count()

        self._clear_events()

        self._test_source.save()

        self.assertEqual(
            PeriodicTask.objects.count(), periodic_task_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_source)
        self.assertEqual(events[0].target, self._test_source)
        self.assertEqual(events[0].verb, event_source_edited.id)


class PeriodicSourceBackendActionTestCase(
    PeriodicSourceBackendTestMixin, GenericDocumentTestCase
):
    auto_upload_test_document = False

    @mock.patch(target='mayan.apps.source_periodic.tests.source_backends.SourceBackendTestPeriodic.action_file_delete')
    def test_dry_run_false(self, mocked_action_file_delete):
        self._silence_logger(name='mayan.apps.converter.backends')

        document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(
            action_name='document_upload', extra_data={'dry_run': False}
        )

        self.assertEqual(mocked_action_file_delete.call_count, 1)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_edited.id)

    @mock.patch(target='mayan.apps.source_periodic.tests.source_backends.SourceBackendTestPeriodic.action_file_delete')
    def test_dry_run_true(self, mocked_action_file_delete):
        self._silence_logger(name='mayan.apps.converter.backends')

        document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(
            action_name='document_upload', extra_data={'dry_run': True}
        )

        self.assertEqual(mocked_action_file_delete.call_count, 0)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_edited.id)

    @mock.patch(target='mayan.apps.source_periodic.tests.source_backends.SourceBackendTestPeriodic.action_file_delete')
    def test_dry_run_missing(self, mocked_action_file_delete):
        self._silence_logger(name='mayan.apps.converter.backends')

        document_count = Document.objects.count()

        self._clear_events()

        self._execute_test_source_action(action_name='document_upload')

        self.assertEqual(mocked_action_file_delete.call_count, 1)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 4)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, test_document_version)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_edited.id)



from .events import event_workflow_template_edited


def method_document_type_workflow_templates_add(self, queryset, user):
    for model_instance in queryset:
        self.workflows.add(model_instance)
        event_workflow_template_edited.commit(
            action_object=self, actor=user, target=model_instance
        )


def method_document_type_workflow_templates_remove(self, queryset, user):
    for model_instance in queryset:
        self.workflows.remove(model_instance)
        event_workflow_template_edited.commit(
            action_object=self, actor=user, target=model_instance
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.backends.classes import DynamicFormModelBackend


class CredentialBackend(DynamicFormModelBackend):
    _backend_app_label = 'credentials'
    _backend_model_name = 'StoredCredential'
    _loader_module_name = 'credential_backends'

    form_fieldsets = (
        (
            _(message='General'), {
                'fields': ('label', 'internal_name')
            }
        ),
    )


# Null backend must be defined here to avoid automatic import.
class CredentialBackendNull(CredentialBackend):
    label = _(message='Null backend')



from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.file_metadata.tests.mixins import DocumentFileMetadataTestMixin

from ..drivers import ClamScanDriver

from .literals import (
    TEST_CLAMSCAN_FILE_METADATA_DOTTED_NAME,
    TEST_CLAMSCAN_FILE_METADATA_VALUE
)


class ClamScanDriverTestCase(
    DocumentFileMetadataTestMixin, GenericDocumentTestCase
):
    _test_document_file_metadata_driver_path = ClamScanDriver.dotted_path

    def test_driver_entries(self):
        self._test_document.submit_for_file_metadata_processing()

        value = self._test_document_file.get_file_metadata(
            dotted_name=TEST_CLAMSCAN_FILE_METADATA_DOTTED_NAME
        )
        self.assertEqual(value, TEST_CLAMSCAN_FILE_METADATA_VALUE)



class DocumentDownloadViewTestMixin:
    def _request_test_document_download_get_view(self):
        return self.get(
            viewname='document_downloads:document_download_single', kwargs={
                'document_id': self._test_document.pk
            }
        )

    def _request_test_document_download_post_view(self):
        data = {}
        for test_document_file, test_document_file_index in enumerate(iterable=self._test_document_file_list):
            data.update(
                {
                    'form-{}-document_file_id'.format(
                        test_document_file_index
                    ): self._test_document_file.pk,
                    'form-{}-include'.format(test_document_file_index): True
                }
            )

        return self.post(
            viewname='document_downloads:document_download_single', kwargs={
                'document_id': self._test_document.pk
            }, data={
                'form-0-document_file_id': self._test_document_file.pk,
                'form-0-include': True,
                'form-TOTAL_FORMS': len(self._test_document_file_list),
                'form-INITIAL_FORMS': '0',
                'form-MAX_NUM_FORMS': ''
            }
        )

    def _request_test_document_multiple_download_get_view(self):
        return self.get(
            viewname='document_downloads:document_download_multiple', query={
                'id_list': ','.join(self._test_document_id_list_string)
            }
        )

    def _request_test_document_multiple_download_post_view(self):
        data = {}
        for test_document_file, test_document_file_index in enumerate(iterable=self._test_document_file_list):
            data.update(
                {
                    'form-{}-document_file_id'.format(
                        test_document_file_index
                    ): self._test_document_file.pk,
                    'form-{}-include'.format(test_document_file_index): True
                }
            )

        return self.post(
            viewname='document_downloads:document_download_multiple', query={
                'id_list': ','.join(self._test_document_id_list_string)
            }, data={
                'form-0-document_file_id': self._test_document_file.pk,
                'form-0-include': True,
                'form-TOTAL_FORMS': len(self._test_document_file_list),
                'form-INITIAL_FORMS': '0',
                'form-MAX_NUM_FORMS': ''
            }
        )


class DocumentFileDownloadAPIViewTestMixin:
    def _request_test_document_file_download_api_view(self):
        return self.get(
            viewname='rest_api:documentfile-download', kwargs={
                'document_id': self._test_document.pk,
                'document_file_id': self._test_document.file_latest.pk
            }
        )


class DocumentFileDownloadViewTestMixin:
    def _request_test_document_file_download_view(self):
        return self.get(
            viewname='document_downloads:document_file_download', kwargs={
                'document_file_id': self._test_document_file.pk
            }
        )



from mayan.apps.testing.tests.base import BaseTestCase

from ..models import EventSubscription, ObjectEventSubscription

from .mixins import (
    EventObjectTestMixin, EventTypeTestMixin, NotificationTestMixin
)


class EventTypeNotificationModelTestCase(
    EventObjectTestMixin, EventTypeTestMixin, NotificationTestMixin,
    BaseTestCase
):
    def setUp(self):
        super().setUp()

        self._create_test_event_type()
        self._create_test_object_with_event_type_and_permission()

        EventSubscription.objects.create(
            stored_event_type=self._test_event_type.get_stored_event_type(),
            user=self._test_case_user
        )

    def test_event_type_notification_creation(self):
        test_notification_count = self._test_case_user.notifications.count()

        self._create_test_event(target=self._test_object)
        self.assertEqual(
            self._test_case_user.notifications.count(),
            test_notification_count + 1
        )


class ObjectEventNotificationModelTestCase(
    EventObjectTestMixin, EventTypeTestMixin, NotificationTestMixin,
    BaseTestCase
):
    def setUp(self):
        super().setUp()

        self._create_test_event_type()
        self._create_test_object_with_event_type_and_permission()

        ObjectEventSubscription.objects.create(
            content_object=self._test_object,
            stored_event_type=self._test_event_type.get_stored_event_type(),
            user=self._test_case_user
        )

    def test_object_notification_creation(self):
        test_notification_count = self._test_case_user.notifications.count()

        self._create_test_event(target=self._test_object)
        self.assertEqual(
            self._test_case_user.notifications.count(),
            test_notification_count + 1
        )



from django.db import migrations


def code_move_content_from_documents_to_ocr_app(apps, schema_editor):
    DocumentPage = apps.get_model(
        app_label='documents', model_name='DocumentPage'
    )
    DocumentPageContent = apps.get_model(
        app_label='ocr', model_name='DocumentPageContent'
    )

    for document_page in DocumentPage.objects.using(alias=schema_editor.connection.alias).all():
        DocumentPageContent.objects.using(alias=schema_editor.connection.alias).create(
            document_page=document_page,
            content=document_page.content_old or ''
        )


class Migration(migrations.Migration):
    dependencies = [
        ('ocr', '0002_documentpagecontent')
    ]
    operations = [
        migrations.RunPython(
            code=code_move_content_from_documents_to_ocr_app
        )
    ]
    run_before = [
        ('documents', '0006_remove_documentpage_content_old')
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_d

queue_ocr = CeleryQueue(name='ocr', label=_(message='OCR'), worker=worker_d)

queue_ocr.add_task_type(
    dotted_path='mayan.apps.ocr.tasks.task_document_version_ocr_finished',
    label=_(message='Finish document version OCR')
)
queue_ocr.add_task_type(
    dotted_path='mayan.apps.ocr.tasks.task_document_version_page_ocr_process',
    label=_(message='Document version page OCR')
)
queue_ocr.add_task_type(
    dotted_path='mayan.apps.ocr.tasks.task_document_version_ocr_process',
    label=_(message='Document version OCR')
)



import datetime
import json
import logging

from django.apps import apps
from django.conf import settings
from django.utils.timezone import now

from mayan.apps.acls.models import AccessControlList

from ..permissions import permission_workflow_instance_transition

logger = logging.getLogger(name=__name__)


class WorkflowInstanceBusinessLogicMixin:
    def check_escalation(self):
        current_state = self.get_current_state()

        for escalation in current_state.escalations.filter(enabled=True):
            timedelta = datetime.timedelta(
                **{
                    escalation.unit: escalation.amount
                }
            )

            if now() > self.get_last_log_entry_datetime() + timedelta:
                condition_context = {'workflow_instance': self}

                if escalation.evaluate_condition(context=condition_context):
                    self.do_transition(
                        transition=escalation.transition,
                        comment=escalation.get_comment()
                    )

    def do_transition(
        self, transition, comment=None, extra_data=None, user=None
    ):
        WorkflowInstanceLogEntry = apps.get_model(
            app_label='document_states', model_name='WorkflowInstanceLogEntry'
        )

        try:
            if transition in self.get_transition_choices(user=user).all():
                if extra_data:
                    context = self.loads()
                    context.update(extra_data)
                    self.dumps(context=context)

                workflow_instance_log_entry = WorkflowInstanceLogEntry(
                    comment=comment or '',
                    extra_data=json.dumps(
                        obj=extra_data or {}
                    ),
                    transition=transition, user=user,
                    workflow_instance=self
                )
                workflow_instance_log_entry._event_actor = user
                workflow_instance_log_entry.save()
                return workflow_instance_log_entry
        except AttributeError:
            # No initial state has been set for this workflow.
            if settings.DEBUG:
                raise

    def dumps(self, context):
        """
        Serialize the context data.
        """
        self.context = json.dumps(obj=context)
        self.save()

    def get_context(self):
        # Keep the document instance in the workflow instance fresh when
        # there are cascade state actions, where a second state action is
        # triggered by the events generated by a first state action.
        self.document.refresh_from_db()
        context = {
            'workflow_instance': self
        }
        context['workflow_instance_context'] = self.loads()
        return context

    def get_current_state(self):
        """
        Actual State - The current state of the workflow. If there are
        multiple states available, for example: registered, approved,
        archived; this field will tell at the current state where the
        document is right now.
        """
        last_transition = self.get_last_transition()
        if last_transition:
            return last_transition.destination_state
        else:
            return self.workflow.get_initial_state()

    def get_last_log_entry(self):
        return self.log_entries.order_by('datetime').last()

    def get_last_log_entry_datetime(self):
        last_log_entry = self.get_last_log_entry()
        if last_log_entry:
            return last_log_entry.datetime
        else:
            return self.datetime

    def get_last_transition(self):
        """
        Last Transition - The last transition used by the last user to put
        the document in the actual state.
        """
        last_log_entry = self.get_last_log_entry()
        if last_log_entry:
            return last_log_entry.transition

    def get_runtime_context(self):
        """
        Alias of self.load() to get just the runtime context of the instance
        for ease of use in the condition template.
        """
        return self.loads()

    def get_transition_choices(self, user=None):
        WorkflowTransition = apps.get_model(
            app_label='document_states', model_name='WorkflowTransition'
        )

        current_state = self.get_current_state()

        if current_state:
            queryset = current_state.origin_transitions.all()

            if user:
                queryset = AccessControlList.objects.restrict_queryset(
                    permission=permission_workflow_instance_transition,
                    queryset=queryset, user=user
                )

            # Remove the transitions with a false return value.
            for entry in queryset:
                if not entry.evaluate_condition(workflow_instance=self):
                    queryset = queryset.exclude(id=entry.pk)

            return queryset
        else:
            """
            This happens when a workflow has no initial state and a document
            whose document type has this workflow is created. We return an
            empty transition queryset.
            """
            return WorkflowTransition.objects.none()

    def loads(self):
        """
        Deserialize the context data.
        """
        return json.loads(s=self.context or '{}')


class WorkflowInstanceLogEntryBusinessLogicMixin:
    def get_extra_data(self):
        WorkflowTransitionField = apps.get_model(
            app_label='document_states', model_name='WorkflowTransitionField'
        )

        result = {}
        for key, value in self.loads().items():
            try:
                field = self.transition.fields.get(name=key)
            except WorkflowTransitionField.DoesNotExist:
                """
                There is a reference for a field that does not exist or
                has been deleted.
                """
            else:
                result[field.label] = value

        return result

    def loads(self):
        """
        Deserialize the context data.
        """
        return json.loads(s=self.extra_data or '{}')



from furl import furl

from django.conf import settings
from django.core import mail
from django.utils.html import strip_tags
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

import mayan
from mayan.apps.templating.classes import Template

from .classes import ModelMailingAction
from .events import event_email_sent
from .utils import split_recipient_list


class UserMailerBusinessLogicMixin:
    def get_connection(self):
        """
        Establishes a reusable connection to the server by loading the
        backend, initializing it, and the using the backend instance to get
        a connection.
        """
        backend_instance = self.get_backend_instance()
        connection_kwargs = backend_instance.get_connection_kwargs()

        return mail.get_connection(
            backend=backend_instance.class_path, **connection_kwargs
        )

    def send(
        self, to, _event_action_object=None, attachments=None, bcc=None,
        body='', cc=None, reply_to=None, subject='', user=None
    ):
        """
        Send a simple email. There is no document or template knowledge.
        attachments is a list of dictionaries with the keys:
        filename, content, and mimetype.
        """
        recipient_list = split_recipient_list(
            recipients=[to]
        )

        if cc:
            cc_list = split_recipient_list(
                recipients=[cc]
            )
        else:
            cc_list = None

        if bcc:
            bcc_list = split_recipient_list(
                recipients=[bcc]
            )
        else:
            bcc_list = None

        if reply_to:
            reply_to_list = split_recipient_list(
                recipients=[reply_to]
            )
        else:
            reply_to_list = None

        backend_data = self.get_backend_data()

        with self.get_connection() as connection:
            email_message = mail.EmailMultiAlternatives(
                bcc=bcc_list, body=strip_tags(body), cc=cc_list,
                connection=connection, from_email=backend_data.get('from'),
                reply_to=reply_to_list, subject=subject,
                to=recipient_list
            )

            for attachment in attachments or ():
                email_message.attach(
                    content=attachment['content'],
                    filename=attachment['filename'],
                    mimetype=attachment['mimetype']
                )

            email_message.attach_alternative(body, 'text/html')

        try:
            email_message.send()

        except Exception as exception:
            self.error_log.create(
                text='{}; {}'.format(
                    exception.__class__.__name__, exception
                )
            )
        else:
            self.error_log.all().delete()

            event_email_sent.commit(
                action_object=_event_action_object, actor=user,
                target=self
            )

    def send_object(
        self, obj, to, as_attachment=False, bcc=None, body='', cc=None,
        object_name=None, organization_installation_url='', reply_to=None,
        subject='', user=None
    ):
        """
        Send an object file using this user mailing profile.
        """
        if as_attachment:
            action_name = 'attachment'
        else:
            action_name = 'link'

        model_mailing_action = ModelMailingAction.get_action_for_model(
            action_name=action_name, model=obj._meta.model
        )

        context_dictionary = {
            'link': furl(organization_installation_url).join(
                obj.get_absolute_url()
            ).tostr(),
            'object': obj,
            'object_name': object_name
        }

        body_template = Template(template_string=body)
        body_html_content = body_template.render(
            context=context_dictionary
        )

        subject_template = Template(template_string=subject)
        subject_text = strip_tags(
            subject_template.render(context=context_dictionary)
        )

        attachments = []
        if as_attachment:
            content_function = import_string(
                dotted_path=model_mailing_action.kwargs[
                    'content_function_dotted_path'
                ]
            )

            mime_type_function = import_string(
                dotted_path=model_mailing_action.kwargs[
                    'mime_type_function_dotted_path'
                ]
            )
            mime_type = mime_type_function(obj=obj)

            with content_function(obj=obj) as file_object:
                attachments.append(
                    {
                        'content': file_object.read(),
                        'filename': str(obj),
                        'mimetype': mime_type
                    }
                )

        return self.send(
            _event_action_object=obj, attachments=attachments, bcc=bcc,
            body=body_html_content, cc=cc, reply_to=reply_to,
            subject=subject_text, to=to, user=user
        )

    def test(self, to, user=None):
        """
        Send a test message to make sure the mailing profile settings are
        correct.
        """
        try:
            self.send(
                subject=_(message='Test email from %s') % mayan.__title__,
                to=to, user=user
            )
        except Exception as exception:
            self.error_log.create(
                text='{}; {}'.format(
                    exception.__class__.__name__, exception
                )
            )
            if settings.DEBUG:
                raise
        else:
            self.error_log.all().delete()



from ..models.document_models import Document
from ..tasks import task_document_type_document_stubs_delete

from .base import GenericDocumentTestCase


class SearchTaskTestCase(GenericDocumentTestCase):
    def test_task_document_type_document_stubs_delete(self):
        self._create_test_document_stub()

        self._test_document_type.document_stub_expiration_interval = 0
        self._test_document_type.save()

        test_document_stub_count = Document.objects.count()

        task_document_type_document_stubs_delete.apply_async().get()

        self.assertEqual(
            Document.objects.count(), test_document_stub_count - 1
        )



from django.db import migrations

from mayan.apps.common.utils import convert_to_internal_name


def code_populate_internal_name(apps, schema_editor):
    DocumentFileDriverEntry = apps.get_model(
        app_label='file_metadata', model_name='DocumentFileDriverEntry'
    )
    FileMetadataEntry = apps.get_model(
        app_label='file_metadata', model_name='FileMetadataEntry'
    )

    for document_file_driver_entry in DocumentFileDriverEntry.objects.using(alias=schema_editor.connection.alias).all():
        for file_metadata_entry in document_file_driver_entry.entries.all():
            internal_name = convert_to_internal_name(
                value=file_metadata_entry.key
            )

            queryset_siblings = FileMetadataEntry.objects.using(
                alias=schema_editor.connection.alias
            ).filter(
                document_file_driver_entry=document_file_driver_entry
            ).exclude(pk=file_metadata_entry.pk)

            queryset_duplicated = queryset_siblings.filter(
                internal_name=internal_name
            )

            if queryset_duplicated.exists():
                internal_name = '{}_{}'.format(
                    internal_name, queryset_duplicated.count()
                )

            file_metadata_entry.internal_name = internal_name
            file_metadata_entry.save()


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0010_add_internal_name')
    ]

    operations = [
        migrations.RunPython(
            code=code_populate_internal_name,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _

from mayan.apps.common.utils import convert_to_internal_name
from mayan.apps.documents.models.document_file_models import DocumentFile
from mayan.apps.documents.models.document_type_models import DocumentType

from .managers import DocumentTypeSettingsManager
from .model_mixins import (
    DocumentFileDriverEntryBusinessLogicMixin,
    StoredDriverBusinessLogicMixin
)


class DocumentFileDriverEntry(
    DocumentFileDriverEntryBusinessLogicMixin, models.Model
):
    driver = models.ForeignKey(
        on_delete=models.CASCADE, related_name='driver_entries',
        to='StoredDriver', verbose_name=_(message='Driver')
    )
    document_file = models.ForeignKey(
        on_delete=models.CASCADE, related_name='file_metadata_drivers',
        to=DocumentFile, verbose_name=_(message='Document file')
    )

    class Meta:
        ordering = ('document_file', 'driver')
        unique_together = ('driver', 'document_file')
        verbose_name = _(message='Document file driver entry')
        verbose_name_plural = _(message='Document file driver entries')

    def __str__(self):
        return str(self.driver)


class DocumentTypeSettings(models.Model):
    """
    Model to store the file metadata settings for a document type.
    """
    document_type = models.OneToOneField(
        on_delete=models.CASCADE, related_name='file_metadata_settings',
        to=DocumentType, unique=True, verbose_name=_(message='Document type')
    )
    auto_process = models.BooleanField(
        default=True, help_text=_(
            'Automatically queue newly created documents for processing.'
        ), verbose_name=_(message='Auto process')
    )

    objects = DocumentTypeSettingsManager()

    class Meta:
        verbose_name = _(message='Document type settings')
        verbose_name_plural = _(message='Document types settings')

    def natural_key(self):
        return self.document_type.natural_key()
    natural_key.dependencies = ['documents.DocumentType']


class FileMetadataEntry(models.Model):
    document_file_driver_entry = models.ForeignKey(
        on_delete=models.CASCADE, related_name='entries',
        to=DocumentFileDriverEntry,
        verbose_name=_(message='Document file driver entry')
    )
    internal_name = models.CharField(
        db_index=True, help_text=_(
            message='Normalized name of the file metadata entry.'
        ), max_length=255, verbose_name=_(message='Internal name')
    )
    key = models.CharField(
        help_text=_(
            message='Name of the file metadata entry as provided by the '
            'driver.'
        ), max_length=255, verbose_name=_(message='Key')
    )
    value = models.TextField(
        blank=True, help_text=_(message='Value of the file metadata entry.'),
        max_length=255, verbose_name=_(message='Value')
    )

    class Meta:
        ordering = ('internal_name', 'value')
        unique_together = ('document_file_driver_entry', 'internal_name')
        verbose_name = _(message='File metadata entry')
        verbose_name_plural = _(message='File metadata entries')

    def __str__(self):
        return '{}: {}: {}'.format(
            self.document_file_driver_entry, self.key, self.value
        )

    def save(self, *args, **kwargs):
        internal_name = convert_to_internal_name(value=self.key)

        queryset_siblings = self.document_file_driver_entry.entries.exclude(
            pk=self.pk
        )

        queryset_duplicated = queryset_siblings.filter(
            internal_name=internal_name
        )

        if queryset_duplicated.exists():
            internal_name = '{}_{}'.format(
                internal_name, queryset_duplicated.count()
            )

        self.internal_name = internal_name

        super().save(*args, **kwargs)


class StoredDriver(StoredDriverBusinessLogicMixin, models.Model):
    driver_path = models.CharField(
        max_length=255, unique=True, verbose_name=_(message='Driver path')
    )
    internal_name = models.CharField(
        db_index=True, max_length=128, unique=True,
        verbose_name=_(message='Internal name')
    )

    class Meta:
        ordering = ('internal_name',)
        verbose_name = _(message='Driver')
        verbose_name_plural = _(message='Drivers')

    def __str__(self):
        return str(self.driver_label)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('file_metadata', '0009_alter_documenttypesettings_auto_process')
    ]

    operations = [
        migrations.AddField(
            model_name='filemetadataentry', name='internal_name',
            field=models.CharField(
                db_index=True, default='', help_text='Normalized name '
                'of the file metadata entry.', max_length=255,
                verbose_name='Internal name'
            ), preserve_default=False
        ),
        migrations.AlterField(
            model_name='filemetadataentry', name='key',
            field=models.CharField(
                help_text='Name of the file metadata entry as provided '
                'by the driver.', max_length=255, verbose_name='Key'
            )
        )
    ]



from django.utils.translation import gettext_lazy as _

from mayan.apps.smart_settings.settings import setting_cluster

from .literals import (
    DEFAULT_AUTHENTICATION_BACKEND, DEFAULT_AUTHENTICATION_BACKEND_ARGUMENTS,
    DEFAULT_AUTHENTICATION_DISABLE_PASSWORD_RESET
)

setting_namespace = setting_cluster.do_namespace_add(
    label=_(message='Authentication'), name='authentication'
)

setting_disable_password_reset = setting_namespace.do_setting_add(
    default=DEFAULT_AUTHENTICATION_DISABLE_PASSWORD_RESET,
    global_name='AUTHENTICATION_DISABLE_PASSWORD_RESET', help_text=_(
        'Remove the "Forgot your password?" link on the login form used to '
        'trigger the password reset.'
    )
)
setting_authentication_backend = setting_namespace.do_setting_add(
    default=DEFAULT_AUTHENTICATION_BACKEND,
    global_name='AUTHENTICATION_BACKEND',
    help_text=_(
        'Dotted path to the backend used to process user authentication.'
    )
)
setting_authentication_backend_arguments = setting_namespace.do_setting_add(
    default=DEFAULT_AUTHENTICATION_BACKEND_ARGUMENTS,
    global_name='AUTHENTICATION_BACKEND_ARGUMENTS',
    help_text=_(message='Arguments for the AUTHENTICATION_BACKEND.')
)



from django.utils.translation import gettext_lazy as _

from mayan.apps.permissions.classes import PermissionNamespace

namespace = PermissionNamespace(
    label=_(message='Task manager'), name='task_manager'
)

permission_task_view = namespace.add_permission(
    label=_(message='View tasks'), name='task_view'
)



TEST_STORED_CREDENTIAL_BACKEND_PATH = 'mayan.apps.credentials.tests.credential_backends.TestCredentialBackend'
TEST_STORED_CREDENTIAL_BACKEND_DATA_FIELDS = {
    'test_field': 'test_value'
}
TEST_STORED_CREDENTIAL_BACKEND_DATA_PASSWORD = 'test_credential_password'
TEST_STORED_CREDENTIAL_BACKEND_DATA_USERNAME = 'test_credential_username'
TEST_STORED_CREDENTIAL_INTERNAL_NAME = 'test_credential'
TEST_STORED_CREDENTIAL_LABEL = 'test credential'
TEST_STORED_CREDENTIAL_LABEL_EDITED = 'test credential edited'



from django.apps import apps
from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import menu_object
from mayan.apps.events.classes import ModelEventType

from .events import event_document_version_exported
from .links import link_document_version_export
from .permissions import permission_document_version_export


class DocumentExportsApp(MayanAppConfig):
    app_namespace = 'document_exports'
    app_url = 'document_exports'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.document_exports'
    verbose_name = _(message='Document exports')

    def ready(self):
        super().ready()

        DocumentVersion = apps.get_model(
            app_label='documents', model_name='DocumentVersion'
        )

        ModelEventType.register(
            model=DocumentVersion, event_types=(
                event_document_version_exported,
            )
        )

        ModelPermission.register(
            model=DocumentVersion, permissions=(
                permission_document_version_export,
            )
        )

        menu_object.bind_links(
            links=(
                link_document_version_export,
            ),
            sources=(DocumentVersion,)
        )



from django.utils.translation import gettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.events.decorators import method_event
from mayan.apps.events.event_managers import EventManagerMethodAfter

from .events import event_tag_attached, event_tag_removed
from .html_widgets import widget_single_tag


class TagBusinessLogicMixin:
    @method_event(
        action_object='self',
        event=event_tag_attached,
        event_manager_class=EventManagerMethodAfter
    )
    def _attach_to(self, document, user=None):
        self._event_actor = user
        self._event_target = document
        self.documents.add(document)

    @method_event(
        action_object='self',
        event=event_tag_removed,
        event_manager_class=EventManagerMethodAfter
    )
    def _remove_from(self, document, user=None):
        self._event_actor = user
        self._event_target = document
        self.documents.remove(document)

    def attach_to(self, document, user):
        return self._attach_to(document=document, user=user)

    def get_document_count(self, user):
        """
        Return the numeric count of documents that have this tag attached.
        The count is filtered by access.
        """
        return self.get_documents(
            permission=permission_document_view, user=user
        ).count()

    def get_documents(self, user, permission):
        """
        Return a filtered queryset documents that have this tag attached.
        """
        queryset = Document.valid.filter(
            pk__in=self.documents.all()
        )

        queryset = AccessControlList.objects.restrict_queryset(
            permission=permission, queryset=queryset,
            user=user
        )

        return queryset

    def get_preview_widget(self):
        return widget_single_tag(tag=self)
    get_preview_widget.short_description = _(message='Preview')

    def remove_from(self, document, user):
        return self._remove_from(document=document, user=user)



from django.utils.translation import gettext_lazy as _

from mayan.apps.authentication.link_conditions import condition_user_is_authenticated
from mayan.apps.navigation.classes import Link, Separator, Text
from mayan.apps.navigation.utils import factory_condition_queryset_access

from .icons import (
    icon_current_user_detail, icon_group_create, icon_group_edit,
    icon_group_list, icon_group_multiple_delete, icon_group_setup,
    icon_group_single_delete, icon_group_user_list, icon_user_create,
    icon_user_edit, icon_user_group_list, icon_user_list,
    icon_user_multiple_delete, icon_user_set_options, icon_user_setup,
    icon_user_single_delete
)
from .link_conditions import condition_user_is_not_super_user
from .permissions import (
    permission_group_create, permission_group_delete, permission_group_edit,
    permission_group_view, permission_user_create, permission_user_delete,
    permission_user_edit, permission_user_view
)
from .utils import get_user_label_text

# Current user

link_current_user_details = Link(
    args='request.user.id',
    condition=condition_user_is_authenticated,
    icon=icon_current_user_detail, text=_(message='User details'),
    view='user_management:user_details'
)

# Group

link_group_create = Link(
    icon=icon_group_create, permission=permission_group_create,
    text=_(message='Create new group'), view='user_management:group_create'
)
link_group_single_delete = Link(
    args='object.id', icon=icon_group_single_delete,
    permission=permission_group_delete, tags='dangerous',
    text=_(message='Delete'), view='user_management:group_single_delete'
)
link_group_multiple_delete = Link(
    icon=icon_group_multiple_delete, tags='dangerous', text=_(message='Delete'),
    view='user_management:group_multiple_delete'
)
link_group_edit = Link(
    args='object.id', icon=icon_group_edit,
    permission=permission_group_edit, text=_(message='Edit'),
    view='user_management:group_edit'
)
link_group_list = Link(
    icon=icon_group_list, text=_(message='Groups'),
    view='user_management:group_list'
)
link_group_user_list = Link(
    args='object.id', icon=icon_group_user_list,
    permission=permission_group_edit, text=_(message='Users'),
    view='user_management:group_members'
)
link_group_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='auth', model_name='Group',
        callback=condition_user_is_not_super_user,
        object_permission=permission_group_view,
        view_permission=permission_group_create
    ), icon=icon_group_setup, text=_(message='Groups'),
    view='user_management:group_list'
)

# User

link_user_create = Link(
    condition=condition_user_is_authenticated, icon=icon_user_create,
    permission=permission_user_create, text=_(message='Create new user'),
    view='user_management:user_create'
)
link_user_single_delete = Link(
    args='object.id', condition=condition_user_is_authenticated,
    icon=icon_user_single_delete, permission=permission_user_delete,
    tags='dangerous', text=_(message='Delete'),
    view='user_management:user_single_delete'
)
link_user_multiple_delete = Link(
    icon=icon_user_multiple_delete, tags='dangerous', text=_(message='Delete'),
    view='user_management:user_multiple_delete'
)
link_user_edit = Link(
    args='object.id', condition=condition_user_is_authenticated,
    icon=icon_user_edit, permission=permission_user_edit, text=_(message='Edit'),
    view='user_management:user_edit'
)
link_user_group_list = Link(
    args='object.id', condition=condition_user_is_authenticated,
    icon=icon_user_group_list, permission=permission_user_edit,
    text=_(message='Groups'), view='user_management:user_groups'
)
link_user_list = Link(
    icon=icon_user_list, text=_(message='Users'), view='user_management:user_list'
)
link_user_set_options = Link(
    args='object.id', condition=condition_user_is_authenticated,
    icon=icon_user_set_options, permission=permission_user_edit,
    text=_(message='User options'), view='user_management:user_options'
)
link_user_setup = Link(
    condition=factory_condition_queryset_access(
        app_label='auth', model_name='User',
        object_permission=permission_user_view,
        view_permission=permission_user_create,
    ), icon=icon_user_setup, text=_(message='Users'),
    view='user_management:user_list'
)

separator_user_label = Separator()

text_user_label = Text(
    html_extra_classes='menu-user-name', text=get_user_label_text
)
